# encoding: utf-8
# Copyright 1999-2022 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from collections import namedtuple, OrderedDict
from copy import deepcopy
from types import MethodType

from ...compat import reduce, six, Iterable
from ...utils import camel_to_underline, is_namedtuple
from ..expr.mixin import merge_data
from ..utils import FieldRole


class PipelineBit(object):
    def __init__(self, step, output):
        self._step = step
        self._output = output


class PipelineStep(object):
    def __init__(self, step_name, param_names=None, output_names=None):
        self._step_name = step_name
        self._ml_uplink = []
        self._param_names = [camel_to_underline(n) for n in param_names] if param_names is not None else []
        self._output_names = [camel_to_underline(n) for n in output_names] if output_names is not None else []
        self._pipeline = None
        self._input_args = []
        self._input_kwargs = dict()

    def __repr__(self):
        return self.__class__.__name__ + '(' + ', '.join('%s=%s' % (p, getattr(self, p)) for p in self._param_names
                                                         if p in dir(self) and getattr(self, p) is not None) + ')'

    def link(self, *args, **kwargs):
        if 'name' in kwargs:
            name = kwargs['name']
            del kwargs['name']
        else:
            name = None
        self._input_args = [PipelineBit(v._final_step, v._final_step._output_names[0]) if isinstance(v, Pipeline) else v for v in args]
        self._input_kwargs = dict((k, PipelineBit(v._final_step, v._final_step._output_names[0]))
                                  if isinstance(v, Pipeline) else v for k, v in six.iteritems(kwargs))
        inputs = set(self._input_args) | set(six.itervalues(self._input_kwargs))
        first_input = next(iter(inputs))
        self._pipeline = first_input._step._pipeline
        self._ml_uplink = [bit._step for bit in inputs]
        self._pipeline._add_step(self, name=name)

        out_tuple = namedtuple('OutTuple', ' '.join(self._output_names))
        return out_tuple(**dict((out, PipelineBit(self, out)) for out in self._output_names))

    def _eval(self, method_name, result_dict, *args, **kwargs):
        if not (args or kwargs):
            args = [result_dict[b._step][b._output] for b in self._input_args]
            kwargs = dict((k, result_dict[b._step][b._output]) for k, b in six.iteritems(self._input_kwargs))
        results = getattr(self, method_name)(*args, **kwargs)

        # check namedtuple
        if is_namedtuple(results):
            result_dict[self] = results._asdict()
        else:
            if not isinstance(results, Iterable):
                results = (results, )
            result_dict[self] = OrderedDict(zip(self._output_names, results))


class GroupedStep(PipelineStep):
    def __init__(self, **kwargs):
        name = kwargs.get('name')
        super(GroupedStep, self).__init__(name, output_names=['output'])

        self._steps = []
        self._step_dict = dict()
        self._locked = False

    def __repr__(self):
        return self.__class__.__name__ + '(steps=[' + ', '.join('("%s", %s)' % step for step in self._steps) + '])'

    def __getitem__(self, item):
        if isinstance(item, six.integer_types):
            return self._steps[item][1]
        else:
            item = camel_to_underline(item)
            return self._step_dict[item]

    def __getattr__(self, item):
        if '__' not in item or item.startswith('_'):
            return object.__getattribute__(self, item)
        else:
            step_name, _ = item.split('__', 1)
            if step_name not in self._step_dict:
                raise AttributeError
            return object.__getattribute__(self, item)

    def __setattr__(self, item, value):
        if '__' not in item or item.startswith('_'):
            object.__setattr__(self, item, value)
        else:
            step_name, _ = item.split('__', 1)
            if step_name not in self._step_dict:
                raise AttributeError
            object.__setattr__(self, item, value)

    def _prop_setter_generator(self, step_obj, prop_name):
        def _setter(self, val):
            setattr(step_obj, prop_name, val)
            return self

        return MethodType(_setter, self)

    @staticmethod
    def _prop_generator(step_obj, prop_name):
        def _getter(self):
            return getattr(step_obj, prop_name)

        def _setter(self, val):
            setattr(step_obj, prop_name, val)

        return property(fget=_getter, fset=_setter)

    def _add_step(self, step, name=None):
        if self._locked:
            raise RuntimeError('This object has already been locked.')
        if isinstance(step, GroupedStep):
            step._locked = True

        if name is None:
            new_name = camel_to_underline(step._step_name)
        else:
            new_name = camel_to_underline(name)

        if step in self._step_dict:
            raise ValueError('Step names must be unique.')
        self._output_names = deepcopy(step._output_names)
        self._steps.append((new_name, step))
        self._step_dict[new_name] = step

        cls = type(self)
        if not hasattr(cls, '__perinstance'):
            cls = type(cls.__name__, (cls,), {})
            cls.__perinstance = True

        for step_prop in step._param_names:
            if step_prop in dir(step):
                prop_name = camel_to_underline(new_name) + '__' + camel_to_underline(step_prop)
                setattr(cls, 'set_' + prop_name, self._prop_setter_generator(step, step_prop))
                setattr(cls, prop_name, self._prop_generator(step, step_prop))
                self._param_names.append(prop_name)

        self.__class__ = cls

        return step


class Pipeline(GroupedStep):
    """
    Pipeline defines a sequence of algorithms that can be treated as one algorithm.

    :Examples:
    >>> # define a name-default algorithm sequence
    >>> Pipeline(Normalize(), LogisticRegression())
    >>> # define a named step in the algorithm sequence
    >>> Pipeline(('norm', Normalize()), LogisticRegression())
    >>> # define an output-specified step in the algorithm sequence
    >>> Pipeline(SplitWord(), (DocWordCount(), 'multi'), Word2Vec())
    """
    def __init__(self, *steps, **kwargs):
        name = kwargs.get('name')
        super(Pipeline, self).__init__(name=name if name is not None else 'pipeline')

        self._initial_step = None
        self._final_step = None

        last_step = last_output = None
        for name, step, output in self._normalize_step_input(steps):
            ret_step = self._add_step(step, name=name)
            ret_step._ml_uplink = [last_step, ] if last_step is not None else []
            ret_step._input_args = [PipelineBit(last_step, last_output or last_step._output_names[0])]\
                if last_step is not None else None
            ret_step._pipeline = self
            last_step = ret_step
            last_output = output

        self._final_output = last_output
        self._step_dict = dict(self._steps)

    @staticmethod
    def _normalize_step_input(steps):
        # merge input arrays
        listed = map(lambda s: [s, ] if isinstance(s, (tuple, PipelineStep)) else s, steps)
        condensed = reduce(lambda a, b: a + b, listed, [])

        def normalize_tuple(tp):
            if isinstance(tp, PipelineStep):
                tp = (tp, )
            output_name = None
            if isinstance(tp[0], PipelineStep):
                step = tp[0]
                step_name = camel_to_underline(step._step_name)
                if len(tp) >= 2:
                    output_name = tp[1]
            else:
                step_name = camel_to_underline(tp[0])
                step = tp[1]
                if len(tp) >= 3:
                    output_name = tp[2]
            return step_name, step, output_name

        return list(map(normalize_tuple, condensed))

    def _add_step(self, step, name=None):
        step = super(Pipeline, self)._add_step(step, name=name)
        if self._initial_step is None:
            self._initial_step = step
        self._final_step = step
        return step

    def _assert_valid(self):
        if not self._steps:
            raise ValueError('Cannot execute without any steps defined.')
        zero_input_nodes = [sobj for sname, sobj in self._steps if not sobj._ml_uplink]
        if len(zero_input_nodes) > 1:
            raise ValueError('Multiple input steps is not allowed.')
        rev_dict = dict((s, []) for _, s in self._steps)
        [rev_dict[ul].append(dl) for _, dl in self._steps for ul in dl._ml_uplink]
        mid_items = [s for s, dl in six.iteritems(rev_dict) if dl]
        last_items = [s for s, dl in six.iteritems(rev_dict) if not dl]
        if len(last_items) > 1:
            raise ValueError('Multiple output steps is not allowed.')
        for mid_item in mid_items:
            if 'transform' not in dir(mid_item):
                raise ValueError('Intermediate steps must have transform() method.')

    def _eval_steps(self, step, method_name, result_dict, args, kwargs):
        # run first step
        if step == self._initial_step:
            # initial step needs original args
            if len(self._steps) == 1:
                # if we only have one step, go directly
                self._initial_step._eval(method_name, result_dict, *args, **kwargs)
            else:
                self._initial_step._eval('transform', result_dict, *args, **kwargs)
        else:
            # pass raw args on
            [self._eval_steps(s, 'transform', result_dict, args, kwargs) for s in step._ml_uplink]
            if step != self._final_step:
                step._eval('transform', result_dict)
            else:
                step._eval(method_name, result_dict)

    def _eval_pl(self, method_name, *args, **kwargs):
        self._assert_valid()
        if method_name not in dir(self._final_step):
            raise ValueError('Final step does not contains a method called ' + method_name + '.')
        result_dict = dict()
        self._eval_steps(self._final_step, method_name, result_dict, args, kwargs)
        ret_val = result_dict[self._final_step]

        ret_type = namedtuple('PipelineResult', list(six.iterkeys(ret_val)))
        ret_val = ret_type(**ret_val)

        if len(ret_val) == 1:
            return ret_val[0]
        else:
            return ret_val

    def _eval(self, method_name, result_dict, *args, **kwargs):
        result = self._eval_pl(method_name, *args, **kwargs)
        if is_namedtuple(result):
            result_dict[self] = result._asdict()
        else:
            if not isinstance(result, Iterable):
                result = (result, )
            result_dict[self] = OrderedDict(zip(self._output_names, result))

    def train(self, df):
        return self._eval_pl('train', df)

    def transform(self, *args, **kwargs):
        return self._eval_pl('transform', *args, **kwargs)


class FeatureUnion(GroupedStep):
    """
    FeatureUnion defines a group of feature transform algorithms whose results are eventually merged into one.

    :Examples:
    >>> FeatureUnion(Normalize(), Standardize())
    """
    def __init__(self, *steps, **kwargs):
        name = kwargs.get('name')
        self._include_roles = kwargs.get('include_roles', [FieldRole.LABEL, FieldRole.WEIGHT])
        super(FeatureUnion, self).__init__(name=name if name is not None else 'feature_union')

        self._append_col_dict = dict()
        self.auto_rename = kwargs.get('auto_rename') or True

        for name, step, output, cols in self._normalize_step_input(steps):
            ret_step = self._add_step(step, name=name)
            ret_step._input_args = None
            ret_step._pipeline = self
            self._append_col_dict[name] = cols

        self._step_dict = dict(self._steps)
        self._param_names.append('auto_rename_col')

    @staticmethod
    def _normalize_step_input(steps):
        # merge input arrays
        listed = map(lambda s: [s, ] if isinstance(s, (tuple, PipelineStep)) else s, steps)
        condensed = reduce(lambda a, b: a + b, listed, [])

        def normalize_tuple(tp):
            if isinstance(tp, PipelineStep):
                tp = (tp, )
            output_name = None
            if isinstance(tp[0], PipelineStep):
                step = tp[0]
                step_name = camel_to_underline(step._step_name)
                if len(tp) >= 2:
                    output_name = tp[1]
            else:
                step_name = camel_to_underline(tp[0])
                step = tp[1]
                if len(tp) >= 3:
                    output_name = tp[2]
            if output_name and '[' in output_name:
                output_name, col_str = output_name.split('[', 1)
                cols = [c.strip() for c in col_str.strip(']').split(',')]
            else:
                cols = None
            return step_name, step, output_name, cols

        return list(map(normalize_tuple, condensed))

    def _assert_valid(self):
        if not self._steps:
            raise ValueError('Cannot execute without any steps defined.')
        for _, step in self._steps:
            if 'transform' not in dir(step):
                raise ValueError('Steps must have transform() method.')

    def _eval_fu(self, method_name, *args, **kwargs):
        self._assert_valid()
        result_dict = dict()
        [step._eval('transform', result_dict, *args, **kwargs) for _, step in self._steps]

        processed = reduce(lambda a, b: [(a[0], v) for v in six.itervalues(a[1])] + [(a[0], v) for v in six.itervalues(b[1])],
                           six.iteritems(result_dict))
        results = []
        role_exported = dict((r, False) for r in self._include_roles)
        for src, result in processed:
            user_inclusive = set(self._append_col_dict[src]) if src in self._append_col_dict else set()
            for r in six.iterkeys(role_exported):
                role_fields = [f.name for f in result._ml_fields if r in f.role]
                if role_fields and not role_exported[r]:
                    user_inclusive.update(role_fields)
                    role_exported[r] = True
            exclude_fields = [f.name for f in result._ml_fields
                              if FieldRole.FEATURE not in f.role and f.name not in user_inclusive]
            if exclude_fields:
                results.append((result, exclude_fields, True))
            else:
                results.append(result)

        return merge_data(*results, auto_rename=self.auto_rename)

    def _eval(self, method_name, result_dict, *args, **kwargs):
        result = self._eval_fu(method_name, *args, **kwargs)
        if is_namedtuple(result):
            result_dict[self] = result._asdict()
        else:
            if not isinstance(result, Iterable):
                result = (result, )
            result_dict[self] = OrderedDict(zip(self._output_names, result))

    def transform(self, *args, **kwargs):
        return self._eval_fu('transform', *args, **kwargs)



# Copyright 1999-2022 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import ctypes
import threading
import multiprocessing
import logging

from odps import options
from odps.tunnel.pdio import TunnelPandasReader, BasePandasWriter

from cupid.errors import SubprocessStreamEOFError

import numpy as np
del np

logger = logging.getLogger(__name__)


class CupidPandasReader(TunnelPandasReader):
    def __init__(self, schema, input_stream, columns=None):
        if isinstance(input_stream, tuple):
            self._refill_data = input_stream
            input_stream = None
        else:
            self._refill_data = None

        super(CupidPandasReader, self).__init__(schema, input_stream, columns=columns)

        self._input_stream = input_stream
        self._table_schema = schema
        self._input_columns = columns
        self._stream_eof = False
        self._closed = False

    def to_forkable(self):
        return type(self)(self._table_schema, self._build_refill_data(), self._input_columns)

    def __repr__(self):
        cls = type(self)
        if self._refill_data is not None:
            return '<%s.%s (slave) at 0x%x>' % (cls.__module__, cls.__name__, id(self))
        else:
            return '<%s.%s at 0x%x>' % (cls.__module__, cls.__name__, id(self))

    def _build_refill_data(self):
        if self._input_stream is None:
            return self._refill_data
        if self._refill_data is not None:
            return self._refill_data

        from multiprocessing.sharedctypes import RawArray
        req_queue = multiprocessing.Queue()
        rep_queue = multiprocessing.Queue()
        buf = RawArray(ctypes.c_char, options.cupid.mp_buffer_size)

        def _mp_thread():
            try:
                while True:
                    req_body = req_queue.get(timeout=60)
                    if req_body is None:
                        return
                    left_size, bound = req_body
                    try:
                        buf[:left_size] = buf[bound - left_size:bound]
                        read_size = self._input_stream.readinto(buf, left_size)
                    except SubprocessStreamEOFError:
                        return
                    rep_queue.put(read_size)
            finally:
                rep_queue.put(-1)
                self.close()

        stream_thread = threading.Thread(target=_mp_thread)
        stream_thread.daemon = True
        stream_thread.start()
        self._refill_data = (buf, req_queue, rep_queue)
        return self._refill_data

    def refill_cache(self):
        if self._refill_data is None:
            return super(CupidPandasReader, self).refill_cache()
        if self._stream_eof or self._closed:
            return 0

        buf, req_queue, rep_queue = self._refill_data
        left_size = self.mem_cache_bound - self.row_mem_ptr
        req_queue.put((left_size, self.mem_cache_bound))
        read_size = rep_queue.get(timeout=60)
        if read_size <= 0:
            self._stream_eof = True
            self.close()
            return 0
        self.reset_positions(buf, read_size + left_size)
        return read_size

    def close(self):
        super(CupidPandasReader, self).close()
        if self._input_stream is None and self._refill_data:
            buf, req_queue, rep_queue = self._refill_data
            req_queue.put(None)
        self._closed = True


class CupidPandasWriter(BasePandasWriter):
    def __init__(self, schema, output_stream):
        super(CupidPandasWriter, self).__init__(schema, output_stream)
        self._stream = output_stream
        self._block_id = None
        self._partition_spec = None
        self._table_schema = schema

    @property
    def block_id(self):
        return self._block_id

    @property
    def partition_spec(self):
        return self._partition_spec

    def write_stream(self, data, length):
        self._stream.write(data, length)

    def close(self):
        super(CupidPandasWriter, self).close()

        # sync by get result
        result = self._stream.result()
        logger.debug('Result fetched on writer close: %s', result)

        self._stream.close()



# Copyright 1999-2022 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import argparse
import base64
import logging
import json
import os
import sys

from mars.deploy.kubernetes.client import KubernetesCluster, new_cluster
from mars.deploy.kubernetes.config import (
    VolumeConfig,
    MarsReplicationConfig,
    MarsSupervisorsConfig,
    MarsWorkersConfig,
)
from mars.deploy.utils import wait_services_ready
from mars.utils import readable_size

from ...utils import build_mars_image_name

logger = logging.getLogger(__name__)

UI_PORT = 50002
NOTEBOOK_PORT = 50003
GS_COORDINATOR_PORT = 63800
GS_COORDINATOR_GATEWAY_PORT = 64800


class DiskDriverVolumeConfig(VolumeConfig):
    def __init__(self, name, mount_path, device_size, renew_interval=3, timeout=120):
        super().__init__(name, mount_path)
        self._device_size = device_size
        self._timeout = timeout
        self._renew_interval = renew_interval

    def build(self):
        return {
            "name": self.name,
            "diskDriverEphemeralDevice": {
                "deviceSize": self._device_size,
                "timeout": self._timeout,
                "renewInterval": self._renew_interval,
            },
        }


class CupidMarsConfigMixin:
    _default_kind = "ReplicationController"

    def __init__(self, *args, **kwargs):
        self._container_port = kwargs.pop("container_port", None)
        super().__init__(*args, **kwargs)

    @property
    def stat_type(self):
        return "cgroup" if os.environ.get("VM_ENGINE_TYPE") != "hyper" else None

    @staticmethod
    def get_local_app_module(mod_name):
        from . import supervisor

        return supervisor.__name__.rsplit(".", 1)[0] + "." + mod_name

    def build_container(self):
        container_config = super().build_container()
        container_config.update({"enableChannel": True, "enableNativeLib": True})
        if self._container_port is not None:
            container_config["ports"] = {
                "containerPort": int(self._container_port),
            }
        return container_config

    def build_template_spec(self):
        spec_config = super().build_template_spec()
        if os.environ.get("VM_ENGINE_TYPE") == "hyper":
            spec_config.update({"restartPolicy": "Never"})
        else:
            spec_config.update({"hostNetwork": True, "restartPolicy": "Never"})
        return spec_config

    def config_readiness_probe(self):
        return None

    def build(self):
        result = super().build()
        with open("pod_%s.json" % self.rc_name, "w") as pod_file:
            pod_file.write(json.dumps(result, indent=2, sort_keys=True))
        return result

    def remove_env(self, name):
        try:
            super().remove_env(name)
        except AttributeError:
            self._envs.pop(name, None)


class CupidMarsSupervisorsConfig(CupidMarsConfigMixin, MarsSupervisorsConfig):
    def __init__(self, *args, **kwargs):
        if os.environ.get("VM_ENGINE_TYPE") == "hyper":
            default_port = UI_PORT
        else:
            default_port = None
        kwargs["web_port"] = kwargs.get("web_port", None) or default_port
        super().__init__(*args, **kwargs)

    def add_default_envs(self):
        super().add_default_envs()
        self.remove_env("MARS_USE_CGROUP_STAT")
        self.add_env("MARS_POD_ROLE", "supervisor")
        if self.stat_type == "cgroup":
            self.add_env("MARS_MEM_USE_CGROUP_STAT", "1")
            self.add_env("MARS_CPU_USE_PROCESS_STAT", "1")


class CupidMarsWorkersConfig(CupidMarsConfigMixin, MarsWorkersConfig):
    def __init__(self, *args, **kwargs):
        if os.environ.get("VM_ENGINE_TYPE") == "hyper":
            kwargs["mount_shm"] = kwargs.get("mount_shm", None) or False
        kwargs["min_cache_mem"] = kwargs.get("min_cache_mem", None) or "10%"
        super().__init__(*args, **kwargs)

    def add_default_envs(self):
        super().add_default_envs()
        self.add_env("MARS_DISABLE_PROC_RECOVER", "1")
        self.add_env("MARS_PLASMA_LIMIT", "95%")
        self.add_env("MARS_POD_ROLE", "worker")
        self.remove_env("MARS_USE_CGROUP_STAT")
        if self.stat_type == "cgroup":
            self.add_env("MARS_CPU_USE_PROCESS_STAT", "1")
            self.add_env("MARS_MEM_USE_CGROUP_STAT", "1")


class CupidMarsNotebooksConfig(CupidMarsConfigMixin, MarsReplicationConfig):
    rc_name = "marsnotebook"

    def __init__(self, *args, **kwargs):
        if os.environ.get("VM_ENGINE_TYPE") == "hyper":
            default_port = NOTEBOOK_PORT
        else:
            default_port = None
        kwargs["service_port"] = kwargs.get("service_port", None) or default_port
        super().__init__(*args, **kwargs)

    def build_container_command(self):
        return [
            "/srv/entrypoint.sh",
            self.get_local_app_module("notebook"),
        ]

    def add_default_envs(self):
        super().add_default_envs()
        self.add_env("MARS_POD_ROLE", "notebook")
        if os.environ.get("VM_ENGINE_TYPE") == "hyper":
            self.add_env("MARS_NOTEBOOK_PORT", self._container_port)


class CupidGSCoordinatorConfig(CupidMarsConfigMixin, MarsReplicationConfig):
    rc_name = "gscoordinator"

    def __init__(self, *args, **kwargs):
        if os.environ.get("VM_ENGINE_TYPE") == "hyper":
            default_port = GS_COORDINATOR_GATEWAY_PORT
        else:
            default_port = None
        kwargs["service_port"] = kwargs.get("service_port", None) or default_port
        self._worker_pod_name_list = kwargs.pop("worker_pod_name_list", None)
        self._worker_pod_ip_list = kwargs.pop("worker_pod_ip_list", None)
        self._port = GS_COORDINATOR_PORT
        self._gateway_port = kwargs["service_port"]
        self._worker_num = kwargs.pop("worker_num", 1)
        super().__init__(*args, **kwargs)

    def build_container_command(self):
        # replace command with coordinator
        coordinator_args = dict(
            port=self._port,
            gateway_port=self._gateway_port,
            worker_pod_name_list=self._worker_pod_name_list,
            worker_pod_ip_list=self._worker_pod_ip_list,
            num_workers=self._worker_num,
        )
        cmd = [
            "/srv/entrypoint.sh",
            self.get_local_app_module("gscoordinator"),
            base64.b64encode(json.dumps(coordinator_args).encode()).decode(),
        ]
        return cmd

    def add_default_envs(self):
        super().add_default_envs()
        self.add_env("MARS_POD_ROLE", "gscoordinator")
        if os.environ.get("VM_ENGINE_TYPE") == "hyper":
            self.add_env("GS_COORDINATOR_GATEWAY_PORT", self._container_port)
            self.add_env("GS_COORDINATOR_PORT", self._container_port)


class CupidKubernetesCluster(KubernetesCluster):
    _supervisor_config_cls = CupidMarsSupervisorsConfig
    _worker_config_cls = CupidMarsWorkersConfig
    _default_service_port = None
    _default_web_port = None

    def __init__(self, *args, **kwargs):
        self._with_notebook = kwargs.pop("with_notebook", False)
        self._notebook_cpu = kwargs.pop("notebook_cpu", None)
        self._notebook_mem = kwargs.pop("notebook_mem", None)

        self._notebook_extra_env = kwargs.get("extra_env", None) or dict()
        self._notebook_extra_env.update(
            kwargs.pop("notebook_extra_env", None) or dict()
        )

        self._with_graphscope = kwargs.pop("with_graphscope", False)
        self._coordinator_extra_env = kwargs.get("extra_env", None) or dict()
        self._coordinator_extra_env.update(
            kwargs.pop("coordinator_extra_env", None) or dict()
        )
        self._coordinator_cpu = kwargs.pop("coordinator_cpu", None)
        self._coordinator_mem = kwargs.pop("coordinator_mem", None)

        self._node_blacklist = set(kwargs.pop("node_blacklist", None) or [])
        self._blacklisted_pods = set()

        kwargs["image"] = build_mars_image_name(kwargs.pop("image", None))
        super().__init__(*args, **kwargs)

    def _create_notebook(self):
        if self._with_notebook:
            notebook_config = CupidMarsNotebooksConfig(
                1,
                image=self._image,
                cpu=self._notebook_cpu,
                memory=self._notebook_mem,
                volumes=self._extra_volumes,
                pre_stop_command=self._pre_stop_command,
            )
            notebook_config.add_simple_envs(self._notebook_extra_env)
            self._core_api.create_namespaced_replication_controller(
                self._namespace, notebook_config.build()
            )

    def _create_graphscope(self):
        if self._with_graphscope:
            # check the worker services are ready
            limits = [self._worker_num]
            selectors = ["mars/service-type=" + CupidMarsWorkersConfig.rc_name]
            wait_services_ready(
                selectors,
                limits,
                lambda sel: self._get_ready_pod_count(sel),
                timeout=self._timeout,
            )
            # get the worker pod name and ip
            worker_pod_name_list, worker_pod_ip_list = self._get_pods_name_and_ip(
                selectors[0]
            )
            coordinator_config = CupidGSCoordinatorConfig(
                1,
                image=self._image,
                cpu=self._coordinator_cpu,
                memory=self._coordinator_mem,
                volumes=self._extra_volumes,
                pre_stop_command=self._pre_stop_command,
                worker_pod_name_list=worker_pod_name_list,
                worker_pod_ip_list=worker_pod_ip_list,
                worker_num=self._worker_num,
            )
            coordinator_config.add_simple_envs(self._coordinator_extra_env)
            self._core_api.create_namespaced_replication_controller(
                self._namespace, coordinator_config.build()
            )

    def _create_services(self):
        super()._create_services()
        self._create_graphscope()
        self._create_notebook()

    def _create_kube_service(self):
        # does not create k8s service as not supported in cupid
        pass

    def _get_web_address(self):
        return None

    def _get_ready_pod_count(self, label_selector):
        if self._node_blacklist:
            query = self._core_api.list_namespaced_pod(
                namespace=self._namespace, label_selector=label_selector
            ).to_dict()
            for el in query["items"]:
                node_name = el.get("spec", {}).get("node_name")
                pod_name = el["metadata"]["name"]

                if pod_name in self._blacklisted_pods:
                    continue
                if node_name and node_name in self._node_blacklist:
                    logger.warning(
                        "Found node %s in blacklist, will terminate pod %s",
                        node_name,
                        pod_name,
                    )
                    try:
                        self._core_api.delete_namespaced_pod(
                            name=pod_name, namespace=self._namespace
                        )
                        self._blacklisted_pods.add(pod_name)
                    except:  # noqa: E722
                        pass
        return super()._get_ready_pod_count(label_selector)

    def _get_pods_name_and_ip(self, label_selector):
        pod_name_list = []
        pod_ip_list = []
        query = self._core_api.list_namespaced_pod(
            namespace=self._namespace, label_selector=label_selector
        ).to_dict()
        for el in query["items"]:
            if el["status"]["reason"] == "CupidStarted":
                pod_name_list.append(el["metadata"]["name"])
                pod_ip_list.append(el["status"]["pod_ip"])
        return pod_name_list, pod_ip_list


class MarsCupidServer(object):
    def __init__(self):
        self.args = None
        self._instance_id = None
        self._kube_url = None
        self._kube_client = None

    def __call__(self, argv=None):
        if argv is None:
            argv = sys.argv[1:]
        return self._main(argv)

    def config_logging(self):
        import logging.config
        import mars

        log_conf = self.args.log_conf or "logging.conf"

        conf_file_paths = [
            "",
            os.path.abspath("."),
            os.path.dirname(os.path.dirname(mars.__file__)),
        ]
        log_configured = False
        for path in conf_file_paths:
            conf_path = os.path.join(path, log_conf) if path else log_conf
            if os.path.exists(conf_path):
                logging.config.fileConfig(conf_path, disable_existing_loggers=False)
                log_configured = True
                break

        if not log_configured:
            log_level = self.args.log_level
            level = getattr(logging, log_level.upper()) if log_level else logging.INFO
            logging.getLogger("mars").setLevel(level)
            logging.basicConfig(format=self.args.log_format)

    def _main(self, argv=None):
        parser = argparse.ArgumentParser(description="Mars Cupid Application")
        parser.add_argument("--log-level", help="log level")
        parser.add_argument("--log-format", help="log format")
        parser.add_argument(
            "--log-conf", help="log config file, logging.conf by default"
        )
        parser.add_argument("encoded_args")
        self.args = parser.parse_args(argv)
        self.config_logging()

        self._kube_url = os.environ["KUBE_API_ADDRESS"].strip('"')
        self._instance_id = os.environ["KUBE_NAMESPACE"].strip('"')

        args_dict = json.loads(base64.b64decode(self.args.encoded_args).decode())
        args_dict["namespace"] = self._instance_id

        if args_dict.get("worker_disk_num") is not None:
            disk_num = args_dict.pop("worker_disk_num")
            disk_size = args_dict.pop("worker_disk_size")
            args_dict["worker_spill_paths"] = [
                DiskDriverVolumeConfig(
                    name="diskdriver-volume%d" % i,
                    mount_path="/diskdriver%d" % i,
                    device_size=readable_size(disk_size, trunc=True).lower(),
                )
                for i in range(disk_num)
            ]

        supervisor_extra_env = args_dict.get("supervisor_extra_env") or dict()
        if args_dict.get("instance_idle_timeout") is not None:
            idle_timeout = args_dict.pop("instance_idle_timeout")
            supervisor_extra_env["MARS_INSTANCE_IDLE_TIMEOUT"] = str(idle_timeout)
        args_dict["supervisor_extra_env"] = supervisor_extra_env

        extra_env = args_dict.get("extra_env") or dict()
        extra_env["KUBE_API_ADDRESS"] = self._kube_url
        extra_env["KUBE_NAMESPACE"] = self._instance_id
        args_dict["extra_env"] = extra_env

        new_cluster(
            self.get_instance_kube(), cluster_cls=CupidKubernetesCluster, **args_dict
        )
        logger.info("Cluster creation finished")

    def get_instance_kube(self):
        from kubernetes import client

        config = client.Configuration()
        config.host = self._kube_url
        return client.ApiClient(config)


main = MarsCupidServer()

if __name__ == "__main__":
    main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2022 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .. import types as _types
from ..config import options
try:
    if not options.force_py:
        from ..src.types_c import BaseRecord

        Record = _types.RecordMeta('Record', (_types.RecordReprMixin, BaseRecord),
                                   {'__doc__': _types.Record.__doc__})
    else:
        Record = _types.Record
except (ImportError, AttributeError):
    if options.force_c:
        raise
    Record = _types.Record



# Copyright 1999-2022 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import asyncio
import json
import logging
import multiprocessing
import os
import shutil
import socket
import tempfile
import time
from typing import AsyncGenerator, List, Optional

from mars.deploy.kubernetes.config import MarsSupervisorsConfig
from mars.services.cluster import NodeRole
from mars.services.cluster.backends import (
    AbstractClusterBackend,
    register_cluster_backend,
)

from ..cupid_service import run_cupid_service, CupidServiceClient

logger = logging.getLogger(__name__)


_abs_methods = dict()
for name, method in vars(AbstractClusterBackend).items():
    if getattr(method, "__isabstractmethod__", False):
        _abs_methods[name] = None

_ClusterBackend = type("_ClusterBackend", (AbstractClusterBackend,), _abs_methods)


@register_cluster_backend
class CupidK8SClusterBackend(_ClusterBackend):
    name = "cupid_k8s"
    _pod_to_port = dict()

    def __init__(self, k8s_config=None, k8s_namespace=None):
        from kubernetes import client

        self._k8s_config = k8s_config

        verify_ssl = bool(int(os.environ.get("KUBE_VERIFY_SSL", "1").strip('"')))
        if not verify_ssl:
            c = client.Configuration()
            c.verify_ssl = False
            client.Configuration.set_default(c)

        self._k8s_namespace = (
            k8s_namespace or os.environ.get("MARS_K8S_POD_NAMESPACE") or "default"
        )
        self._full_label_selector = None
        self._client = client.CoreV1Api(client.ApiClient(self._k8s_config))

        self._service_pod_to_ep = dict()

    @classmethod
    async def create(
        cls, node_role: NodeRole, lookup_address: Optional[str], pool_address: str
    ) -> "AbstractClusterBackend":
        from kubernetes import config, client

        if os.environ.get("KUBE_API_ADDRESS"):
            k8s_config = client.Configuration()
            k8s_config.host = os.environ["KUBE_API_ADDRESS"]
            return cls(k8s_config, None)

        if lookup_address is None:
            k8s_namespace = None
            k8s_config = config.load_incluster_config()
        else:
            address_parts = lookup_address.rsplit("?", 1)
            k8s_namespace = None if len(address_parts) == 1 else address_parts[1]

            k8s_config = client.Configuration()
            if "://" in address_parts[0]:
                k8s_config.host = address_parts[0]
            else:
                config.load_kube_config(
                    address_parts[0], client_configuration=k8s_config
                )
        return cls(k8s_config, k8s_namespace)

    def __reduce__(self):
        return type(self), (self._k8s_config, self._k8s_namespace)

    @property
    def cupid_client(self):
        if not hasattr(self, "_cupid_client"):
            self._cupid_client = CupidServiceClient()
        return self._cupid_client

    async def _get_label_selector(self, service_type):
        if self._full_label_selector is not None:
            return self._full_label_selector

        selectors = [f"mars/service-type={service_type}"]
        if "MARS_K8S_GROUP_LABELS" in os.environ:
            group_labels = os.environ["MARS_K8S_GROUP_LABELS"].split(",")
            cur_pod_info = (
                await asyncio.to_thread(
                    self._client.read_namespaced_pod,
                    os.environ["MARS_K8S_POD_NAME"],
                    namespace=self._k8s_namespace,
                )
            ).to_dict()
            for label in group_labels:
                label_val = cur_pod_info["metadata"]["labels"][label]
                selectors.append(f"{label}={label_val}")
        self._full_label_selector = ",".join(selectors)
        logger.debug("Using pod selector %s", self._full_label_selector)
        return self._full_label_selector

    async def _extract_pod_name_ep(self, pod_data):
        pod_name = pod_data["metadata"]["name"]

        if self._pod_to_port.get(pod_name):
            pod_port = self._pod_to_port[pod_name]
        else:
            try:
                pod_kv_data = await asyncio.to_thread(
                    self.cupid_client.get_kv, pod_name
                )
            except OSError:
                logger.warning("Cupid server not ready.")
                return pod_name, None

            if pod_kv_data:
                pod_port = self._pod_to_port[pod_name] = json.loads(pod_kv_data)[
                    "endpoint"
                ].rsplit(":", 1)[-1]
                logger.debug(
                    "Get port from kvstore, name: {}, port: {}".format(
                        pod_name, pod_port
                    )
                )
            else:
                pod_port = None
                if pod_name not in self._pod_to_port:
                    logger.debug(
                        "Cannot get port from kvstore, name: {}".format(pod_name)
                    )
                    self._pod_to_port[pod_name] = None
        pod_endpoint = "%s:%s" % (pod_data["status"]["pod_ip"], pod_port)
        return pod_name, pod_endpoint if pod_port else None

    @staticmethod
    def _extract_pod_ready(obj_data):
        return obj_data["status"]["phase"] == "Running"

    async def _get_pod_to_ep(self, service_type: str, filter_ready: bool = False):
        query = (
            await asyncio.to_thread(
                self._client.list_namespaced_pod,
                namespace=self._k8s_namespace,
                label_selector=await self._get_label_selector(service_type),
                resource_version="0",
            )
        ).to_dict()

        result = dict()
        for el in query["items"]:
            name, pod_ep = await self._extract_pod_name_ep(el)
            if filter_ready and pod_ep is not None and not self._extract_pod_ready(el):
                pod_ep = None
            result[name] = pod_ep
        return result

    async def _get_endpoints_by_service_type(
        self, service_type: str, update: bool = False, filter_ready: bool = True
    ):
        if not self._service_pod_to_ep.get(service_type) or update:
            self._service_pod_to_ep[service_type] = await self._get_pod_to_ep(
                service_type, filter_ready=filter_ready
            )
        return sorted(
            a for a in self._service_pod_to_ep[service_type].values() if a is not None
        )

    async def _watch_service(self, service_type, linger=60):
        from urllib3.exceptions import ReadTimeoutError
        from kubernetes import watch

        cur_pods = set(
            await self._get_endpoints_by_service_type(service_type, update=True)
        )
        w = watch.Watch()

        pod_to_ep = self._service_pod_to_ep[service_type]
        while True:
            # when some pods are not ready, we refresh faster
            linger_seconds = linger() if callable(linger) else linger
            streamer = w.stream(
                self._client.list_namespaced_pod,
                namespace=self._k8s_namespace,
                label_selector=await self._get_label_selector(service_type),
                timeout_seconds=linger_seconds,
                resource_version="0",
            )
            while True:
                try:
                    event = await asyncio.to_thread(next, streamer, StopIteration)
                    if event is StopIteration:
                        # todo change this into a continuous watch
                        #  when watching in a master node is implemented
                        return
                except (ReadTimeoutError, StopIteration):
                    new_pods = set(
                        await self._get_endpoints_by_service_type(
                            service_type, update=True
                        )
                    )
                    if new_pods != cur_pods:
                        cur_pods = new_pods
                        yield await self._get_endpoints_by_service_type(
                            service_type, update=False
                        )
                    break
                except:  # noqa: E722  # pragma: no cover  # pylint: disable=bare-except
                    logger.exception("Unexpected error when watching on kubernetes")
                    break

                obj_dict = event["object"].to_dict()
                pod_name, endpoint = await self._extract_pod_name_ep(obj_dict)
                pod_to_ep[pod_name] = (
                    endpoint if endpoint and self._extract_pod_ready(obj_dict) else None
                )
                yield await self._get_endpoints_by_service_type(
                    service_type, update=False
                )

    def watch_supervisors(self) -> AsyncGenerator[List[str], None]:
        return self._watch_service(MarsSupervisorsConfig.rc_name)

    async def get_supervisors(self, filter_ready: bool = True) -> List[str]:
        return await self._get_endpoints_by_service_type(
            MarsSupervisorsConfig.rc_name,
            update=not filter_ready,
            filter_ready=filter_ready,
        )


class CupidCommandRunnerMixin:
    def fix_hyper_address(self):
        if os.environ.get("VM_ENGINE_TYPE") == "hyper":
            os.environ["MARS_CONTAINER_IP"] = socket.gethostbyname(socket.gethostname())

    def fix_protobuf_import(self):
        # The reason is that plasma and tensorflow relies on protobuf 3+,
        # meanwhile, cupid channel relies on protobuf 2.4,
        # however, when cupid channel started below,
        # tensorflow will recognize the old version of protobuf
        # even when we set LD_LIBRARY_PATH,
        # so we import tensorflow in advance to prevent from potential crash.
        from pyarrow import plasma

        try:
            import tensorflow
        except ImportError:
            tensorflow = None
        del plasma, tensorflow

    def start_cupid_service(self):
        self._env_path = tempfile.mkdtemp(prefix="mars-pool-")
        self._channel_file = os.path.join(
            self._env_path, "mars-cupid-channel-%s.json" % os.getpid()
        )
        self._cupid_sock_file = os.environ["CUPID_SERVICE_SOCKET"] = os.path.join(
            self._env_path, "mars-cupid-sock-%s.sock" % os.getpid()
        )

        self._cupid_service_proc = multiprocessing.Process(
            target=run_cupid_service, args=(self._channel_file,)
        )
        self._cupid_service_proc.start()

        from cupid import context

        self._cupid_context = context()

        envs = self._cupid_context.prepare_channel()
        envs_dict = {env.name: env.value for env in envs}
        with open(self._channel_file, "w") as env_file:
            env_file.write(json.dumps(envs_dict))

        while not os.path.exists(self._cupid_sock_file):
            time.sleep(0.1)

    def stop_cupid_service(self):
        shutil.rmtree(self._env_path)
        if self._cupid_context is not None:
            self._cupid_context.channel_client.stop()
        if self._cupid_service_proc is not None:
            self._cupid_service_proc.terminate()

    def _register_application(self, app_name, internal_endpoint, app_endpoint):
        kvstore = self._cupid_context.kv_store()
        kvstore[app_name] = json.dumps(dict(endpoint=internal_endpoint))
        self._cupid_context.register_application(app_name, app_endpoint)

    async def register_application(self, app_name, internal_endpoint, app_endpoint):
        await asyncio.to_thread(
            self._register_application, app_name, internal_endpoint, app_endpoint
        )

    async def write_node_endpoint(self):
        cupid_client = CupidServiceClient(self._cupid_sock_file)
        content = json.dumps(dict(endpoint=self.args.endpoint))
        await asyncio.to_thread(
            cupid_client.put_kv, os.environ["MARS_K8S_POD_NAME"], content
        )

    async def start_readiness_server(self):
        pass

    async def stop_readiness_server(self):
        pass

    def _get_logging_config_paths(self):
        paths = super()._get_logging_config_paths()
        # make config file inside the package as the highest priority
        config_dir = os.path.dirname(self.get_default_config_file())
        return [os.path.join(config_dir, "logging.conf")] + paths



# Copyright 1999-2022 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import json

__all__ = ['Counter', 'CounterGroup', 'Counters']


class Counter(object):
    def __init__(self, name, value=0):
        self.name = name
        self.value = value
    
    def get_name(self):
        return self.name

    def get_value(self):
        return self.value

    def set_value(self, value):
        self.value = value

    def increment(self, incr):
        self.value += incr

    def _get_data_obj(self):
        data = {
                'name' : self.name,
                'value' : self.value
                }

        return data


class CounterGroup(object):
    def __init__(self, name):
        self.name = name
        self.counters = dict()

    def get_name(self):
        return self.name

    def add_counter(self, counter):
        self.counters[counter.get_name()] = counter

    def get_counter(self, counter_name):
        return self.counters.setdefault(counter_name, Counter(counter_name))

    def size(self):
        return len(self.counters)

    def _get_data_obj(self):
        data = {
                'name' : self.name,
                'counters' : [c._get_data_obj() for c in self.counters.values()]
                }

        return data


class Counters(object):
    def __init__(self):
        self.groups = dict()

    def get_group(self, group_name):
        return self.groups.setdefault(group_name, CounterGroup(group_name))

    def get_counter(self, group_name, counter_name):
        return self.get_group(group_name).get_counter(counter_name)

    def size(self):
        return len(self.groups)

    def to_json_string(self, **json_options):
        data = dict()
        for (k, v) in self.groups.items():
            data[k] = v._get_data_obj()

        return json.dumps(data, **json_options)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2022 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import
import inspect
import functools
import operator
import sys
from collections import defaultdict, OrderedDict

from .core import Node, NodeMetaclass
from .errors import ExpressionError
from .utils import get_attrs, is_called_by_inspector, highest_precedence_data_type, new_id, \
    is_changed, get_proxied_expr
from .. import types
from ...compat import reduce, isvalidattr, dir2, lkeys, six, futures, Iterable
from ...config import options
from ...errors import NoSuchObject, DependencyNotInstalledError
from ...utils import TEMP_TABLE_PREFIX, to_binary, deprecated, survey
from ...models import TableSchema


class ReprWrapper(object):
    def __init__(self, func, repr):
        self._func = func
        self._repr = repr
        functools.update_wrapper(self, func)

    def __call__(self, *args, **kwargs):
        return self._func(*args, **kwargs)

    def __repr__(self):
        return self._repr(self._func)


def _wrap_method_repr(func):
    def inner(*args, **kwargs):
        obj = func(*args, **kwargs)
        if inspect.ismethod(obj):
            def _repr(x):
                instance = getattr(x, 'im_self', getattr(x, '__self__', None))
                method = 'bound method' if instance is not None else 'unbound method'
                if instance is not None:
                    return '<%(method)s %(instance)s.%(name)s>' % {
                        'method': method,
                        'instance': getattr(instance, 'node_name', instance.__class__.__name__),
                        'name': x.__name__
                    }
                else:
                    return '<function __main__.%s>' % x.__name__

            return ReprWrapper(obj, _repr)
        return obj

    return inner


def make_max_pt_function(expr=None):
    from ...models import Table
    from .. import func

    def max_pt(table_name=None):
        if isinstance(getattr(expr, 'data', None), Table):
            table_name = table_name or expr.data.name
        return func.max_pt(table_name)

    return max_pt


def repr_obj(obj):
    if hasattr(obj, '_repr'):
        try:
            return obj._repr()
        except:
            return object.__repr__(obj)
    elif isinstance(obj, (tuple, list)):
        return ','.join(repr_obj(it) for it in obj)

    return obj


class Expr(Node):
    __slots__ = '_deps', '_ban_optimize', '_engine', '_need_cache', '_mem_cache', '__execution', '_id'

    def _init(self, *args, **kwargs):
        """
        _deps is used for common dependencies.
        When a expr depend on other exprs, and the expr is not calculated from the others,
        the _deps are specified to identify the dependencies.
        """
        self._init_attr('_deps', None)

        self._init_attr('_ban_optimize', False)
        self._init_attr('_engine', None)
        self._init_attr('_Expr__execution', None)

        self._init_attr('_need_cache', False)
        self._init_attr('_mem_cache', False)

        if '_id' not in kwargs:
            kwargs['_id'] = new_id()

        super(Expr, self)._init(*args, **kwargs)

    def __repr__(self):
        if not options.interactive or is_called_by_inspector():
            return self._repr()
        else:
            if isinstance(self.__execution, Exception):
                self.__execution = None
            if self.__execution is None:
                try:
                    self.__execution = self.execute()
                except Exception as e:
                    self.__execution = e
                    raise
            return self.__execution.__repr__()

    def _repr_html_(self):
        if not options.interactive:
            return '<code>' + repr(self) + '</code>'
        else:
            if self.__execution is None:
                self.__execution = self.execute()
            else:
                if isinstance(self.__execution, Exception):
                    try:
                        return
                    finally:
                        self.__execution = None
            if hasattr(self.__execution, '_repr_html_'):
                return self.__execution._repr_html_()
            return repr(self.__execution)

    def _handle_delay_call(self, method, *args, **kwargs):
        delay = kwargs.pop('delay', None)
        if delay is not None:
            future = delay.register_item(method, *args, **kwargs)
            return future
        else:
            from ..engines import get_default_engine
            engine = get_default_engine(self)

            wrapper = kwargs.pop('wrapper', None)
            result = getattr(engine, method)(*args, **kwargs)
            if wrapper is None:
                return result
            async_ = kwargs.get('async_', kwargs.get('async', False))
            if async_:
                user_future = futures.Future()

                def _relay(f):
                    try:
                        user_future.set_result(wrapper(f.result()))
                    except:
                        if hasattr(f, 'exception_info'):
                            user_future.set_exception_info(*f.exception_info())
                        else:
                            user_future.set_exception(f.exception())

                result.add_done_callback(_relay)
                return user_future
            else:
                return wrapper(result)

    def visualize(self):
        from ..engines import get_default_engine

        engine = get_default_engine(self)
        return engine.visualize(self)

    def execute(self, **kwargs):
        """
        :param hints: settings for SQL, e.g. `odps.sql.mapper.split.size`
        :type hints: dict
        :param priority: instance priority, 9 as default
        :type priority: int
        :param running_cluster: cluster to run this instance
        :return: execution result
        :rtype: :class:`odps.df.backends.frame.ResultFrame`
        """
        _wrapper = kwargs.pop('wrapper', None)

        def wrapper(result):
            self.__execution = result
            if _wrapper is not None:
                return _wrapper(result)
            else:
                return result

        return self._handle_delay_call('execute', self, wrapper=wrapper, **kwargs)

    def compile(self):
        """
        Compile this expression into an ODPS SQL

        :return: compiled DAG
        :rtype: str
        """

        from ..engines import get_default_engine

        engine = get_default_engine(self)
        return engine.compile(self)

    def persist(self, name, partitions=None, partition=None, lifecycle=None, project=None, **kwargs):
        """
        Persist the execution into a new table. If `partitions` not specified,
        will create a new table without partitions if the table does not exist,
        and insert the SQL result into it.
        If `partitions` are specified, they will be the partition fields of the new table.
        If `partition` is specified, the data will be inserted into the exact partition of the table.

        :param name: table name
        :param partitions: list of string, the partition fields
        :type partitions: list
        :param partition: persist to a specified partition
        :type partition: string or PartitionSpec
        :param lifecycle: table lifecycle. If absent, `options.lifecycle` will be used.
        :type lifecycle: int
        :param project: project name, if not provided, will be the default project
        :param hints: settings for SQL, e.g. `odps.sql.mapper.split.size`
        :type hints: dict
        :param priority: instance priority, 9 as default
        :type priority: int
        :param running_cluster: cluster to run this instance
        :param overwrite: overwrite the table, True as default
        :type overwrite: bool
        :param drop_table: drop table if exists, False as default
        :type drop_table: bool
        :param create_table: create table first if not exits, True as default
        :type create_table: bool
        :param drop_partition: drop partition if exists, False as default
        :type drop_partition: bool
        :param create_partition: create partition if not exists, None as default
        :type create_partition: bool
        :param cast: cast all columns' types as the existed table, False as default
        :type cast: bool
        :return: :class:`odps.df.DataFrame`

        :Example:

        >>> df = df['name', 'id', 'ds']
        >>> df.persist('odps_new_table')
        >>> df.persist('odps_new_table', partition='pt=test')
        >>> df.persist('odps_new_table', partitions=['ds'])
        """
        if lifecycle is None and options.lifecycle is not None:
            lifecycle = \
                options.lifecycle if not name.startswith(TEMP_TABLE_PREFIX) \
                    else options.temp_lifecycle

        return self._handle_delay_call('persist', self, name, partitions=partitions, partition=partition,
                                       lifecycle=lifecycle, project=project, **kwargs)

    def cache(self, mem=False):
        self._need_cache = True
        self._mem_cache = mem
        self._ban_optimize = True
        return self

    def uncache(self):
        self._need_cache = False
        self._mem_cache = False

        from ..backends.context import context
        context.uncache(self)

    def verify(self):
        """
        Verify if this expression can be compiled into ODPS SQL.

        :return: True if compilation succeed else False
        :rtype: bool
        """

        try:
            self.compile()

            return True
        except:
            return False

    def _repr(self):
        from .formatter import ExprFormatter

        formatter = ExprFormatter(self)
        return formatter()

    def ast(self):
        """
        Return the AST string.

        :return: AST tree
        :rtype: str
        """

        return self._repr()

    @_wrap_method_repr
    def __getattribute__(self, attr):
        try:
            return super(Expr, self).__getattribute__(attr)
        except AttributeError:
            if not attr.startswith('_'):
                new_attr = '_%s' % attr
                if new_attr in object.__getattribute__(self, '_args_indexes'):
                    try:
                        return object.__getattribute__(self, new_attr)
                    except AttributeError:
                        return

            raise

    def _defunc(self, field):
        return field(self) if inspect.isfunction(field) else field

    @property
    def optimize_banned(self):
        return self._ban_optimize

    @optimize_banned.setter
    def optimize_banned(self, val):
        self._ban_optimize = val

    @property
    def args(self):
        if not self._deps:
            return super(Expr, self).args
        return super(Expr, self).args + self.deps

    @property
    def deps(self):
        if self._deps is None:
            return
        return tuple(dep if not isinstance(dep, tuple) else dep[0]
                     for dep in self._deps)

    def add_deps(self, *deps):
        dependencies = []
        if len(deps) == 1 and isinstance(deps[0], Iterable):
            dependencies.append([d for d in dependencies
                                 if isinstance(d, Expr)])
        else:
            dependencies.extend(deps)

        if getattr(self, '_deps', None) is None:
            self._deps = dependencies
        else:
            self._deps.extend(dependencies)

    def substitute(self, old_arg, new_arg, dag=None):
        super(Expr, self).substitute(old_arg, new_arg, dag=dag)

        new_deps = []
        if self._deps is not None:
            for dep in self._deps:
                if isinstance(dep, tuple):
                    node = dep[0]
                    if node is old_arg:
                        new_deps.append((new_arg, ) + dep[1:])
                    else:
                        new_deps.append(dep)
                else:
                    if dep is old_arg:
                        new_deps.append(new_arg)
                    else:
                        new_deps.append(dep)

        self._deps = new_deps

    def rebuild(self):
        # used in the dynamic setting, do nothing by default
        # `rebuild` will copy itself, and apply all changes to the new one
        return self.copy()

    def __hash__(self):
        return self._node_id * hash(Expr)

    def __eq__(self, other):
        try:
            return self._eq(other)
        except AttributeError:
            # Due to current complexity of parent's eq,
            # by now, every expression is unequal
            return self is other

    def __ne__(self, other):
        try:
            return self._ne(other)
        except AttributeError:
            return not super(Expr, self).__eq__(other)

    def __lt__(self, other):
        return self._lt(other)

    def __le__(self, other):
        return self._le(other)

    def __gt__(self, other):
        return self._gt(other)

    def __ge__(self, other):
        return self._ge(other)

    def __add__(self, other):
        return self._add(other)

    def __radd__(self, other):
        return self._radd(other)

    def __mul__(self, other):
        return self._mul(other)

    def __rmul__(self, other):
        return self._rmul(other)

    def __div__(self, other):
        return self._div(other)

    def __rdiv__(self, other):
        return self._rdiv(other)

    __truediv__ = __div__
    __rtruediv__ = __rdiv__

    def __floordiv__(self, other):
        return self._floordiv(other)

    def __rfloordiv__(self, other):
        return self._rfloordiv(other)

    def __mod__(self, other):
        return self._mod(other)

    def __rmod__(self, other):
        return self._rmod(other)

    def __sub__(self, other):
        return self._sub(other)

    def __rsub__(self, other):
        return self._rsub(other)

    def __pow__(self, power):
        return self._pow(power)

    def __rpow__(self, power):
        return self._rpow(power)

    def __or__(self, other):
        return self._or(other)

    def __ror__(self, other):
        return self._ror(other)

    def __and__(self, other):
        return self._and(other)

    def __rand__(self, other):
        return self._rand(other)

    def __neg__(self):
        return self._neg()

    def __invert__(self):
        return self._invert()

    def __abs__(self):
        return self._abs()


class CollectionExpr(Expr):
    """
    Collection represents for the two-dimensions data.

    :Example:

    >>> # projection
    >>> df = DataFrame(o.get_table('my_table')) # DataFrame is actually a CollectionExpr
    >>> df['name', 'id']  # projection some columns
    >>> df[[df.name, df.id]]  # projection
    >>> df[df]  # means nothing, but get all the columns
    >>> df[df, df.name.lower().rename('name2')]  # projection a new columns `name2` besides all the original columns
    >>> df.select(df, name2=df.name.lower())  # projection by `select`
    >>> df.exclude('name')  # projection all columns but `name`
    >>> df[df.exclude('name'), df.name.lower()]  # `name` will not conflict any more
    >>>
    >>> # filter
    >>> df[(df.id < 3) & (df.name != 'test')]
    >>> df.filter(df.id < 3, df.name != 'test')
    >>>
    >>> # slice
    >>> df[: 10]
    >>> df.limit(10)
    >>>
    >>> # Sequence
    >>> df.name # an instance of :class:`odps.df.expr.expressions.SequenceExpr`
    >>>
    >>> # schema or dtypes
    >>> df.dtypes
    odps.Schema {
      name    string
      id      int64
    }
    >>> df.schema
    odps.Schema {
      name    string
      id      int64
    }
    """

    __slots__ = (
        '_schema', '_source_data', '_proxy',
        '_ml_fields_cache', '_ml_uplink', '_ml_operations',
    )
    node_name = 'Collection'

    def _init(self, *args, **kwargs):
        self._init_attr('_source_data', None)
        self._init_attr('_proxy', None)

        self._init_attr('_ml_fields_cache', None)
        self._init_attr('_ml_uplink', [])
        self._init_attr('_ml_operations', [])

        super(CollectionExpr, self)._init(*args, **kwargs)

        if hasattr(self, '_schema') and any(it is None for it in self._schema.names):
            raise TypeError('Schema cannot has field whose name is None')

    def __dir__(self):
        dir_set = set(dir2(self)) | set([c.name for c in self.schema if isvalidattr(c.name)])
        return sorted(dir_set)

    def __getitem__(self, item):
        item = self._defunc(item)
        if isinstance(item, tuple):
            item = list(item)
        if isinstance(item, CollectionExpr):
            item = [item, ]

        if isinstance(item, six.string_types):
            return self._get_field(item)
        elif isinstance(item, slice):
            if item.start is None and item.stop is None and item.step is None:
                return self
            return self._slice(item)
        elif isinstance(item, list):
            return self._project(item)
        else:
            field = self._get_field(item)
            if isinstance(field, BooleanSequenceExpr):
                return self.filter(item)

        if isinstance(item, SequenceExpr):
            raise ExpressionError('No boolean sequence found for filtering, '
                                  'a tuple or list is required for projection')
        raise ExpressionError('Not supported projection: collection[%s]' % repr_obj(item))

    def _set_field(self, value, value_dag=None):
        expr = self.copy() if self._proxy is None else self._proxy._input

        if value_dag is None:
            value_dag = value.to_dag(copy=False, validate=False)

        if value_dag.contains_node(self):
            value_dag.substitute(self, expr)
            value = value_dag.root

        if self._proxy is not None:
            self._proxy._setitem(value, value_dag=value_dag)
            return

        fields = [f if f != value.name else value for f in self._schema.names]
        if value.name not in self._schema:
            fields.append(value)
        # make the isinstance to check the proxy type
        self.__class__ = type(self.__class__.__name__, (self.__class__,), {})
        self._proxy = expr.select(fields)

    def __setitem__(self, key, value):
        if not isinstance(value, Expr) and value is not None:
            value = Scalar(value)

        if not isinstance(key, tuple):
            if value is None:
                raise ValueError('Cannot determine type for column %s with None value.' % key)
            column_name = key
            value = value.rename(column_name)
        else:
            conds = key[:-1]
            conds = [self._defunc(c) for c in conds]
            if not all(isinstance(c.dtype, types.Boolean) for c in conds):
                raise ValueError('Conditions should be boolean expressions or boolean columns')
            if len(conds) == 1:
                cond = conds[0]
            else:
                cond = reduce(operator.and_, conds)
            column_name = key[-1]
            if column_name not in self._schema:
                if value is None:
                    raise ValueError('Cannot determine type for column %s with None value.' % key)
                default_col = Scalar(_value_type=value.dtype)
            else:
                default_col = self[column_name]
                if value is None:
                    value = Scalar(_value_type=default_col.dtype)
            value = cond.ifelse(value, default_col).rename(column_name)
        self._set_field(value)

    def __delitem__(self, key):
        if key not in self.schema:
            raise KeyError('Field({0}) does not exist'.format(key))

        if self._proxy is not None:
            return self._proxy._delitem(key)

        fields = [n for n in self._schema.names if n != key]
        # make the instance to check the proxy type
        self.__class__ = type(self.__class__.__name__, (self.__class__,), {})
        self._proxy = self.copy().select(fields)

    def __delattr__(self, item):
        if item in self._schema:
            return self.__delitem__(item)

        super(CollectionExpr, self).__delattr__(item)

    def __setattr__(self, key, value):
        try:
            if key != '_proxy' and object.__getattribute__(self, '_proxy'):
                return setattr(self._proxy, key, value)
        except AttributeError:
            pass

        Expr.__setattr__(self, key, value)

    def __getattribute__(self, attr):
        try:
            proxy = object.__getattribute__(self, '_proxy')
            if attr == '_proxy':
                return proxy

            if proxy:
                # delegate everything to proxy object
                return getattr(proxy, attr)
        except AttributeError:
            pass

        try:
            if attr in object.__getattribute__(self, '_schema')._name_indexes:
                cls_attr = getattr(type(self), attr, None)
                if cls_attr is None or inspect.ismethod(cls_attr) or inspect.isfunction(cls_attr):
                    return self[attr]
        except AttributeError:
            pass

        return super(CollectionExpr, self).__getattribute__(attr)

    def query(self, expr):
        """
        Query the data with a boolean expression.

        :param expr: the query string, you can use '@' character refer to environment variables.
        :return: new collection
        :rtype: :class:`odps.df.expr.expressions.CollectionExpr`

        """
        from .query import CollectionVisitor

        if not isinstance(expr, six.string_types):
            raise ValueError('expr must be a string')

        frame = sys._getframe(2).f_locals
        try:
            env = frame.copy()
        finally:
            del frame

        env['max_pt'] = ReprWrapper(make_max_pt_function(self), repr)

        visitor = CollectionVisitor(self, env)
        predicate = visitor.eval(expr)
        return self.filter(predicate)

    def filter(self, *predicates):
        """
        Filter the data by predicates

        :param predicates: the conditions to filter
        :return: new collection
        :rtype: :class:`odps.df.expr.expressions.CollectionExpr`
        """
        predicates = self._get_fields(predicates)

        predicate = reduce(operator.and_, predicates)
        return FilterCollectionExpr(self, predicate, _schema=self._schema)

    @deprecated('The function `filter_partition` is deprecated, please use `filter_parts` instead'
                "and change the predicate parameter into `pt1=1,pt2=2/pt1=2,pt2=1` form.")
    @survey
    def filter_partition(self, predicate='', exclude=True):
        if isinstance(predicate, six.string_types):
            part_reprs = '/'.join(','.join(p.split('/')) for p in predicate.split(','))
        else:
            part_reprs = predicate
        return self.filter_parts(part_reprs, exclude)

    def filter_parts(self, predicate='', exclude=True):
        """
        Filter the data by partition string. A partition string looks like `pt1=1,pt2=2/pt1=2,pt2=1`, where
        comma (,) denotes 'and', while (/) denotes 'or'.

        :param str|Partition predicate: predicate string of partition filter
        :param bool exclude: True if you want to exclude partition fields, otherwise False. True for default.
        :return: new collection
        :rtype: :class:`odps.df.expr.expressions.CollectionExpr`
        """
        source = self._source_data
        if source is None:
            raise ExpressionError('Can only filter on data sources.')

        def _parse_partition_predicate(p):
            if '=' not in p:
                raise ExpressionError('Illegal partition predicate.')
            field_name, field_value = [s.strip() for s in p.split('=', 1)]
            if not hasattr(source, 'table_schema'):
                raise ExpressionError('filter_partition can only be applied on ODPS DataFrames')
            if field_name not in source.table_schema:
                raise ExpressionError('Column `%s` not exists in input collection' % field_name)
            if field_name not in source.table_schema._partition_schema:
                raise ExpressionError('`%s` is not a partition column' % field_name)
            part_col = self[field_name]
            if field_value.startswith('\'') or field_value.startswith('\"'):
                encoding = 'string-escape' if six.PY2 else 'unicode-escape'
                field_value = to_binary(field_value.strip('"\'')).decode(encoding)

            if isinstance(part_col.data_type, types.Integer):
                field_value = int(field_value)
            elif isinstance(part_col.data_type, types.Float):
                field_value = float(field_value)
            return part_col == field_value

        from ...models.partition import Partition
        from ...types import PartitionSpec

        if isinstance(predicate, Partition):
            predicate = predicate.partition_spec
        if isinstance(predicate, PartitionSpec):
            predicate = ','.join("%s='%s'" % (k, v) for k, v in six.iteritems(predicate.kv))

        if isinstance(predicate, list):
            predicate = '/'.join(str(s) for s in predicate)
        elif not isinstance(predicate, six.string_types):
            raise ExpressionError('Only accept string predicates.')

        if not predicate:
            predicate_obj = None
        else:
            part_formatter = lambda p: reduce(operator.and_, map(_parse_partition_predicate, p.split(',')))
            predicate_obj = reduce(operator.or_, map(part_formatter, predicate.split('/')))

        if not source.table_schema.partitions:
            raise ExpressionError('No partition columns in the collection.')
        if exclude:
            columns = [
                c for c in self.schema if c.name not in source.table_schema._partition_schema
            ]
            new_schema = types.TableSchema.from_lists([c.name for c in columns], [c.type for c in columns])
            return FilterPartitionCollectionExpr(self, predicate_obj, _schema=new_schema, _predicate_string=predicate)
        else:
            return self.filter(predicate_obj)

    def _validate_field(self, field):
        if not isinstance(field, SequenceExpr):
            return True

        if not field.is_ancestor(self):
            return False

        for path in field.all_path(self):
            if any(isinstance(n, CollectionExpr) for n in path[1: -1]):
                return False

            from .reduction import GroupedSequenceReduction
            if any(isinstance(n, GroupedSequenceReduction) for n in path):
                return False
        return True

    @staticmethod
    def _backtrack_field(field, collection):
        from .window import RankOp

        for col in field.traverse(top_down=True, unique=True,
                                  stop_cond=lambda x: isinstance(x, (Column, RankOp))):
            if isinstance(col, Column):
                changed = is_changed(collection, col)
                if changed or changed is None:
                    return
                elif col.source_name not in collection.schema:
                    return
                else:
                    col.substitute(col._input, collection)
            elif isinstance(col, RankOp) and col._input is not collection:
                col.substitute(col._input, collection)

        return field

    def _get_field(self, field):
        from .reduction import GroupedSequenceReduction

        field = self._defunc(field)

        if isinstance(field, six.string_types):
            if field not in self._schema:
                raise ValueError('Field(%s) does not exist, please check schema' % field)
            cls = Column
            if callable(getattr(type(self), field, None)):
                cls = CallableColumn
            return cls(self, _name=field, _data_type=self._schema[field].type)

        if not self._validate_field(field):
            new_field = None
            if not isinstance(field, GroupedSequenceReduction):
                # the reduction is not allowed
                new_field = self._backtrack_field(field, self)
            if new_field is None:
                raise ExpressionError('Cannot support projection on %s' % repr_obj(field))
            field = new_field
        return field

    def _get_fields(self, fields, ret_raw_fields=False):
        selects = []
        raw_selects = []

        for field in fields:
            field = self._defunc(field)
            if isinstance(field, CollectionExpr):
                if any(c is self for c in field.children()):
                    selects.extend(self._get_fields(field._project_fields))
                else:
                    selects.extend(self._get_fields(field._fetch_fields()))
                raw_selects.append(field)
            else:
                select = self._get_field(field)
                selects.append(select)
                raw_selects.append(select)

        if ret_raw_fields:
            return selects, raw_selects
        return selects

    @classmethod
    def _backtrack_lateral_view(cls, lv, collection):
        if isinstance(lv.input, ProjectCollectionExpr):
            src_inputs = (lv.input, lv.input.input)
        else:
            src_inputs = (lv.input,)
        lv_copy = lv.copy(_id=None, _lateral_view=True)

        cur = collection
        while cur not in src_inputs and isinstance(cur, (ProjectCollectionExpr, FilterCollectionExpr)):
            cur = cur.input
        if cur not in src_inputs:
            raise ExpressionError("Input of 'apply' in lateral views can only be "
                                  "simple column selections")
        if cur is lv.input:
            apply_src = lv_copy
        else:
            apply_src = lv_copy.input

        if collection is apply_src.input:
            return lv_copy

        for f in apply_src._fields:
            cls._backtrack_field(f, collection)
        lv_copy.substitute(apply_src.input, collection)
        return lv_copy

    def _filter_lateral_views(self, fields):
        from .collections import RowAppliedCollectionExpr
        from .merge import JoinFieldMergedCollectionExpr, JoinCollectionExpr, UnionCollectionExpr

        walk_through_collections = (JoinCollectionExpr, UnionCollectionExpr,
                                    JoinFieldMergedCollectionExpr)

        lateral_views = []
        for field in fields:
            if not isinstance(field, RowAppliedCollectionExpr):
                continue
            else:
                is_lv = True
                stop_cond = lambda n: n is not self and isinstance(n, CollectionExpr) \
                                      and not isinstance(n, walk_through_collections)
                for coll in self.traverse(unique=True, stop_cond=stop_cond):
                    if coll is field:
                        is_lv = False
                        break
                if is_lv:
                    lateral_views.append(field)
        return lateral_views

    def _project(self, fields):
        field_lvs = self._filter_lateral_views(fields)
        lv_id_set = set(id(f) for f in field_lvs)

        selects = []
        lateral_views = []
        for idx, field in enumerate(fields):
            if id(field) not in lv_id_set:
                s = self._get_fields([field])
                selects.extend(s)
            else:
                lv = self._backtrack_lateral_view(field, self)
                lateral_views.append(lv)
                selects.extend(lv.columns)

        names = [f.name for f in selects]
        typos = [f.dtype for f in selects]

        if not lateral_views and all(isinstance(it, Scalar) for it in selects):
            return self._summary(selects)

        if len(names) != len(set(names)):
            counts = defaultdict(lambda: 0)
            for n in names:
                counts[n] += 1
            raise ExpressionError('Duplicate column names: %s' %
                                  ', '.join(n for n in counts if counts[n] > 1))

        if lateral_views:
            return LateralViewCollectionExpr(self, _fields=selects, _lateral_views=lateral_views,
                                             _schema=types.TableSchema.from_lists(names, typos))
        else:
            return ProjectCollectionExpr(self, _fields=selects,
                                         _schema=types.TableSchema.from_lists(names, typos))

    def select(self, *fields, **kw):
        """
        Projection columns. Remember to avoid column names' conflict.

        :param fields: columns to project
        :param kw: columns and their names to project
        :return: new collection
        :rtype: :class:`odps.df.expr.expression.CollectionExpr`
        """
        if len(fields) == 1 and isinstance(fields[0], list):
            fields = fields[0]
        else:
            fields = list(fields)
        if kw:
            def handle(it):
                it = self._defunc(it)
                if not isinstance(it, Expr):
                    it = Scalar(it)
                return it
            fields.extend([handle(f).rename(new_name)
                           for new_name, f in six.iteritems(kw)])

        return self._project(fields)

    def exclude(self, *fields):
        """
        Projection columns which not included in the fields

        :param fields: field names
        :return: new collection
        :rtype: :class:`odps.df.expr.expression.CollectionExpr`
        """

        if len(fields) == 1 and isinstance(fields[0], list):
            exclude_fields = fields[0]
        else:
            exclude_fields = list(fields)

        exclude_fields = [self._defunc(it) for it in exclude_fields]
        exclude_fields = [field.name if not isinstance(field, six.string_types) else field
                          for field in exclude_fields]

        fields = [name for name in self._schema.names
                  if name not in exclude_fields]

        return self._project(fields)

    def _summary(self, fields):
        names = [field if isinstance(field, six.string_types) else field.name
                 for field in fields]
        typos = [self._schema.get_type(field) if isinstance(field, six.string_types)
                 else field.dtype for field in fields]
        if None in names:
            raise ExpressionError('Column does not have a name, '
                                  'please specify one by `rename`')
        return Summary(_input=self, _fields=fields,
                       _schema=types.TableSchema.from_lists(names, typos))

    def _slice(self, slices):
        return SliceCollectionExpr(self, _indexes=slices, _schema=self._schema)

    @property
    def schema(self):
        return self._schema

    @property
    def odps_schema(self):
        from ..backends.odpssql.types import df_schema_to_odps_schema
        return df_schema_to_odps_schema(self.schema)

    @property
    def columns(self):
        """
        :return: columns
        :rtype: list which each element is a Column
        """

        return [self[n] for n in self._schema.names]

    def _fetch_fields(self):
        return [self._get_field(name) for name in self._schema.names]

    @property
    def _project_fields(self):
        return self._fetch_fields()

    def _data_source(self):
        if hasattr(self, '_source_data') and self._source_data is not None:
            yield self._source_data

    def __getattr__(self, attr):
        try:
            obj = super(CollectionExpr, self).__getattribute__(attr)

            return obj
        except AttributeError as e:
            if attr in object.__getattribute__(self, '_schema')._name_indexes:
                return self[attr]

            raise e

    def output_type(self):
        return 'collection'

    def limit(self, n):
        """
        limit n records

        :param n: n records
        :return:
        """

        return self[:n]

    def head(self, n=None, **kwargs):
        """
        Return the first n rows. Execute at once.

        :param n:
        :return: result frame
        :rtype: :class:`odps.df.backends.frame.ResultFrame`
        """
        if n is None:
            n = options.display.max_rows
        return self._handle_delay_call('execute', self, head=n, **kwargs)

    def tail(self, n=None, **kwargs):
        """
        Return the last n rows. Execute at once.

        :param n:
        :return: result frame
        :rtype: :class:`odps.df.backends.frame.ResultFrame`
        """
        if n is None:
            n = options.display.max_rows
        return self._handle_delay_call('execute', self, tail=n, **kwargs)

    def to_pandas(self, wrap=False, **kwargs):
        """
        Convert to pandas DataFrame. Execute at once.

        :param wrap: if True, wrap the pandas DataFrame into a PyODPS DataFrame
        :return: pandas DataFrame
        """

        try:
            import pandas as pd
        except ImportError:
            raise DependencyNotInstalledError(
                    'to_pandas requires `pandas` library')

        def wrapper(result):
            res = result.values
            if wrap:
                from .. import DataFrame
                return DataFrame(res, schema=self.schema)
            return res

        return self.execute(wrapper=wrapper, **kwargs)

    @property
    def dtypes(self):
        return self.schema

    def view(self):
        """
        Clone a same collection. useful for self-join.

        :return:
        """
        proxied = get_proxied_expr(self)
        kv = dict((attr, getattr(proxied, attr)) for attr in get_attrs(proxied))
        return type(proxied)(**kv)

    def describe(self):
        from .. import output

        numeric_methods = OrderedDict([
            ('count', lambda f: f.count().rename(col.name + '_count')),
            ('mean', lambda f: f.mean()),
            ('std', lambda f: f.std()),
            ('min', lambda f: f.min()),
            ('quantile_25', lambda f: f.quantile(0.25).rename(col.name + '_quantile_25')),
            ('quantile_50', lambda f: f.quantile(0.5).rename(col.name + '_quantile_50')),
            ('quantile_75', lambda f: f.quantile(0.75).rename(col.name + '_quantile_75')),
            ('max', lambda f: f.max()),
        ])
        string_methods = OrderedDict([
            ('count', lambda f: f.count().rename(col.name + '_count')),
            ('unique', lambda f: f.nunique().rename(col.name + '_unique')),
        ])

        aggs = []
        output_names = []
        output_types = []

        def produce_stat_fields(col, methods):
            output_names.append(col.name)
            fields = []
            tps = []
            for func in six.itervalues(methods):
                field = func(self[col.name])
                fields.append(field)
                tps.append(field.dtype)
            t = highest_precedence_data_type(*tps)
            output_types.append(t)
            aggs.extend([f.astype(t) if t == types.decimal else f for f in fields])

        if any(types.is_number(col.type) for col in self.schema.columns):
            methods = numeric_methods
            for col in self.schema.columns:
                if types.is_number(col.type):
                    produce_stat_fields(col, numeric_methods)
        else:
            methods = string_methods
            for col in self.schema.columns:
                produce_stat_fields(col, string_methods)

        summary = self[aggs]
        methods_names = lkeys(methods)

        @output(['type'] + output_names, ['string'] + output_types)
        def to_methods(row):
            for method in methods_names:
                values = [None] * len(output_names)
                for i, field_name in enumerate(output_names):
                    values[i] = getattr(row, '%s_%s' % (field_name, method.split('(', 1)[0]))
                yield [method, ] + values

        return summary.apply(to_methods, axis=1)

    def accept(self, visitor):
        if self._source_data is not None:
            visitor.visit_source_collection(self)
        else:
            raise NotImplementedError

    def get_cached(self, data):
        try:
            if hasattr(data, 'reload'):
                data.reload()
        except NoSuchObject:
            from ..backends.context import context
            context.uncache(self)
            return None

        coll = CollectionExpr(_source_data=data, _schema=self._schema)
        for attr in CollectionExpr.__slots__:
            if not attr.startswith('_ml_'):
                continue
            if hasattr(self, attr):
                setattr(coll, attr, getattr(self, attr))
        return coll


_cached_typed_expr = dict()


class TypedExpr(Expr):
    __slots__ = '_name', '_source_name'

    @classmethod
    def _get_type(cls, *args, **kwargs):
        # return the data type which extracted from args and kwargs
        raise NotImplementedError

    @classmethod
    def _typed_classes(cls, *args, **kwargs):
        # return allowed data types
        raise NotImplementedError

    @classmethod
    def _base_class(cls, *args, **kwargs):
        # base class, SequenceExpr or Scalar
        raise NotImplementedError

    @classmethod
    def _new_cls(cls, *args, **kwargs):
        data_type = cls._get_type(*args, **kwargs)
        if data_type:
            base_class = cls._base_class(*args, **kwargs)
            typed_classes = cls._typed_classes(*args, **kwargs)

            data_type = types.validate_data_type(data_type)
            name = data_type.CLASS_NAME + base_class.__name__
            # get the typed class, e.g. Int64SequenceExpr, StringScalar
            typed_cls = globals().get(name)
            if typed_cls is None:
                raise TypeError("Name {} doesn't exist".format(name))

            if issubclass(cls, typed_cls):
                return cls
            elif cls == base_class:
                return typed_cls
            elif cls in typed_classes:
                return typed_cls

            keys = (cls, typed_cls)
            if keys in _cached_typed_expr:
                return _cached_typed_expr[keys]

            mros = inspect.getmro(cls)
            has_data_type = len([sub for sub in mros if sub in typed_classes]) > 0
            if has_data_type:
                mros = mros[1:]
            subs = [sub for sub in mros if sub not in typed_classes]
            subs.insert(1, typed_cls)

            bases = []
            for sub in subs[::-1]:
                for i in range(len(bases)):
                    if bases[i] is None:
                        continue
                    if issubclass(sub, bases[i]):
                        bases[i] = None
                bases.append(sub)
            bases = tuple(base for base in bases if base is not None)

            dic = dict()
            if hasattr(cls, '_args'):
                dic['_args'] = cls._args
            dic['__slots__'] = cls.__slots__ + getattr(cls, '_slots', ())
            dic['_add_args_slots'] = True
            try:
                accept_cls = next(c for c in bases if c.__name__ == cls.__name__)
                if hasattr(accept_cls, 'accept'):
                    dic['accept'] = accept_cls.accept
            except StopIteration:
                pass
            clz = type(cls.__name__, bases, dic)
            _cached_typed_expr[keys] = clz
            return clz
        else:
            return cls

    def __new__(cls, *args, **kwargs):
        clz = cls._new_cls(*args, **kwargs)
        return super(TypedExpr, clz).__new__(clz)

    def __getattribute__(self, item):
        try:
            return super(TypedExpr, self).__getattribute__(item)
        except AttributeError:
            pass

        # need to show type of expression when certain method missing
        dtype = None
        try:
            dtype = object.__getattribute__(self, "_data_type")
        except AttributeError:  # pragma: no cover
            pass
        try:
            dtype = object.__getattribute__(self, "_value_type")
        except AttributeError:  # pragma: no cover
            pass

        if dtype is not None:
            raise AttributeError(
                "'%s' object has no attribute '%s'. Type of the expression "
                "is %s which may not support this method." % (type(self).__name__, item, dtype)
            )
        else:
            raise AttributeError(
                "'%s' object has no attribute '%s'" % (type(self).__name__, item)
            )

    @classmethod
    def _new(cls, *args, **kwargs):
        return cls._new_cls(*args, **kwargs)(*args, **kwargs)

    def is_renamed(self):
        return self._name is not None and self._source_name is not None and \
               self._name != self._source_name

    def rename(self, new_name):
        if new_name == self._name:
            return self

        attr_dict = dict((attr, getattr(self, attr, None)) for attr in get_attrs(self))
        attr_dict['_source_name'] = self._source_name
        attr_dict['_name'] = new_name

        return type(self)(**attr_dict)

    @property
    def name(self):
        return self._name

    @property
    def source_name(self):
        return self._source_name

    def astype(self, data_type):
        raise NotImplementedError

    def cast(self, t):
        return self.astype(t)

    def eval(self, str_expr, rewrite=False):
        from .query import SequenceVisitor

        if not isinstance(str_expr, six.string_types):
            raise ValueError('expr must be a string')

        frame = sys._getframe(2).f_locals
        try:
            env = frame.copy()
        finally:
            del frame

        env['max_pt'] = ReprWrapper(make_max_pt_function(self), repr)

        visitor = SequenceVisitor(self, env)
        return visitor.eval(str_expr, rewrite=rewrite)


class SequenceExpr(TypedExpr):
    """
    Sequence represents for 1-dimension data.
    """

    __slots__ = (
        '_data_type', '_source_data_type',
        '_ml_fields_cache', '_ml_uplink', '_ml_operations',
    )

    @classmethod
    def _get_type(cls, *args, **kwargs):
        return types.validate_data_type(kwargs.get('_data_type'))

    @classmethod
    def _typed_classes(cls, *args, **kwargs):
        return _typed_sequence_exprs

    @classmethod
    def _base_class(cls, *args, **kwargs):
        return SequenceExpr

    def _init(self, *args, **kwargs):
        self._init_attr('_name', None)

        self._init_attr('_ml_fields_cache', None)
        self._init_attr('_ml_uplink', [])
        self._init_attr('_ml_operations', [])
        super(SequenceExpr, self)._init(*args, **kwargs)

        if '_data_type' in kwargs:
            self._data_type = types.validate_data_type(kwargs.get('_data_type'))

        if '_source_name' not in kwargs:
            self._source_name = self._name

        if '_source_data_type' in kwargs:
            self._source_data_type = types.validate_data_type(kwargs.get('_source_data_type'))
        else:
            self._source_data_type = self._data_type

    def cache(self, mem=False):
        raise ExpressionError('Cache operation does not support for sequence.')

    def head(self, n=None, **kwargs):
        """
        Return first n rows. Execute at once.

        :param n:
        :return: result frame
        :rtype: :class:`odps.df.expr.expressions.CollectionExpr`
        """
        if n is None:
            n = options.display.max_rows
        return self._handle_delay_call('execute', self, head=n, **kwargs)

    def tail(self, n=None, **kwargs):
        """
        Return the last n rows. Execute at once.

        :param n:
        :return:
        """

        if n is None:
            n = options.display.max_rows
        return self._handle_delay_call('execute', self, tail=n, **kwargs)

    def to_pandas(self, wrap=False, **kwargs):
        """
        Convert to pandas Series. Execute at once.

        :param wrap: if True, wrap the pandas DataFrame into a PyODPS DataFrame
        :return: pandas Series
        """

        try:
            import pandas as pd
        except ImportError:
            raise DependencyNotInstalledError(
                    'to_pandas requires for `pandas` library')

        def wrapper(result):
            df = result.values
            if wrap:
                from .. import DataFrame
                df = DataFrame(df)
            return df[self.name]

        return self.execute(wrapper=wrapper, **kwargs)

    @property
    def data_type(self):
        return self._data_type

    @property
    def source_data_type(self):
        return self._source_data_type

    @property
    def dtype(self):
        """
        Return the data type. Available types:
        int8, int16, int32, int64, float32, float64, boolean, string, decimal, datetime

        :return: the data type
        """
        return self._data_type

    def astype(self, data_type):
        """
        Cast to a new data type.

        :param data_type: the new data type
        :return: casted sequence

        :Example:

        >>> df.id.astype('float')
        """

        data_type = types.validate_data_type(data_type)

        if data_type == self._data_type:
            return self

        attr_dict = dict()
        attr_dict['_data_type'] = data_type
        attr_dict['_source_data_type'] = self._source_data_type
        attr_dict['_input'] = self

        new_sequence = AsTypedSequenceExpr(**attr_dict)

        return new_sequence

    def output_type(self):
        return 'sequence(%s)' % repr(self._data_type)

    def accept(self, visitor):
        visitor.visit_sequence(self)


class BooleanSequenceExpr(SequenceExpr):
    def _init(self, *args, **kwargs):
        super(BooleanSequenceExpr, self)._init(*args, **kwargs)
        self._data_type = types.boolean


class Int8SequenceExpr(SequenceExpr):
    def _init(self, *args, **kwargs):
        super(Int8SequenceExpr, self)._init(*args, **kwargs)
        self._data_type = types.int8


class Int16SequenceExpr(SequenceExpr):
    def _init(self, *args, **kwargs):
        super(Int16SequenceExpr, self)._init(*args, **kwargs)
        self._data_type = types.int16


class Int32SequenceExpr(SequenceExpr):
    def _init(self, *args, **kwargs):
        super(Int32SequenceExpr, self)._init(*args, **kwargs)
        self._data_type = types.int32


class Int64SequenceExpr(SequenceExpr):
    def _init(self, *args, **kwargs):
        super(Int64SequenceExpr, self)._init(*args, **kwargs)
        self._data_type = types.int64


class Float32SequenceExpr(SequenceExpr):
    def _init(self, *args, **kwargs):
        super(Float32SequenceExpr, self)._init(*args, **kwargs)
        self._data_type = types.float32


class Float64SequenceExpr(SequenceExpr):
    def _init(self, *args, **kwargs):
        super(Float64SequenceExpr, self)._init(*args, **kwargs)
        self._data_type = types.float64


class DecimalSequenceExpr(SequenceExpr):
    def _init(self, *args, **kwargs):
        super(DecimalSequenceExpr, self)._init(*args, **kwargs)
        self._data_type = types.decimal


class StringSequenceExpr(SequenceExpr):
    def _init(self, *args, **kwargs):
        super(StringSequenceExpr, self)._init(*args, **kwargs)
        self._data_type = types.string


class DatetimeSequenceExpr(SequenceExpr):
    def _init(self, *args, **kwargs):
        super(DatetimeSequenceExpr, self)._init(*args, **kwargs)
        self._data_type = types.datetime


class BinarySequenceExpr(SequenceExpr):
    def _init(self, *args, **kwargs):
        super(BinarySequenceExpr, self)._init(*args, **kwargs)
        self._data_type = types.binary


class DateSequenceExpr(SequenceExpr):
    def _init(self, *args, **kwargs):
        super(DateSequenceExpr, self)._init(*args, **kwargs)
        self._data_type = types.date


class TimestampSequenceExpr(SequenceExpr):
    def _init(self, *args, **kwargs):
        super(TimestampSequenceExpr, self)._init(*args, **kwargs)
        self._data_type = types.timestamp


class JsonSequenceExpr(SequenceExpr):
    def _init(self, *args, **kwargs):
        super(JsonSequenceExpr, self)._init(*args, **kwargs)
        self._data_type = types.json


class ListSequenceExpr(SequenceExpr):
    def _init(self, *args, **kwargs):
        super(ListSequenceExpr, self)._init(*args, **kwargs)


class DictSequenceExpr(SequenceExpr):
    def _init(self, *args, **kwargs):
        super(DictSequenceExpr, self)._init(*args, **kwargs)


class StructSequenceExpr(SequenceExpr):
    def _init(self, *args, **kwargs):
        super(StructSequenceExpr, self)._init(*args, **kwargs)


class UnknownSequenceExpr(SequenceExpr):
    def _init(self, *args, **kwargs):
        super(UnknownSequenceExpr, self)._init(*args, **kwargs)
        if not isinstance(self._data_type, types.Unknown):
            self._data_type = types.Unknown()


_typed_sequence_exprs = [globals()[t.__class__.__name__ + SequenceExpr.__name__]
                         for t in types._data_types.values()]

number_sequences = [globals().get(repr(t).capitalize() + SequenceExpr.__name__)
                    for t in types.number_types()]

int_number_sequences = [globals().get(repr(t).capitalize() + SequenceExpr.__name__)
                        for t in types.number_types() if repr(t).startswith('int')]


class AsTypedSequenceExpr(SequenceExpr):
    _args = '_input',
    node_name = "TypedSequence"

    @property
    def input(self):
        return self._input

    def accept(self, visitor):
        return visitor.visit_cast(self)

    @property
    def name(self):
        return self._name or self._input.name

    @property
    def source_name(self):
        return self._source_name or self._input.source_name

    @property
    def dtype(self):
        return self._data_type or self._input.data_type

    @property
    def source_type(self):
        return self._source_data_type or self._input._source_data_type

    def rebuild(self):
        attr_dict = self._attr_dict()
        tp = self._copy_type()
        if isinstance(attr_dict['_source_data_type'], types.Unknown):
            attr_dict['_source_data_type'] = self._input.dtype
        return tp(**attr_dict)


class Column(SequenceExpr):
    _args = '_input',

    @property
    def input(self):
        return self._input

    def rebuild(self):
        attr_dict = self._attr_dict()
        tp = self._copy_type()
        new_col = self.input[self.source_name]
        for attr in ('_source_data_type', '_data_type'):
            if isinstance(attr_dict[attr], types.Unknown):
                attr_dict[attr] = new_col.dtype
        return tp(**attr_dict)

    def accept(self, visitor):
        return visitor.visit_column(self)


class CallableColumn(Column):
    def __call__(self, *args, **kwargs):
        return getattr(type(self._input), self.source_name)(self._input, *args, **kwargs)


class Scalar(TypedExpr):
    """
    Represent for the scalar type.

    :param _value: value of the scalar
    :param _value_type: value type of the scalar

    :Example:
    >>> df[df, Scalar(4).rename('append_const')]
    """

    __slots__ = '_value', '_value_type', '_source_value_type'

    @classmethod
    def _get_type(cls, *args, **kwargs):
        value = args[0] if len(args) > 0 else None
        value_type = args[1] if len(args) > 1 else None

        val = kwargs.get('_value')
        if val is None:
            val = value

        value_type = kwargs.get('_value_type', None) or value_type

        if val is None and value_type is None:
            raise ValueError('Either value or value_type should be provided')

        if val is not None and not isinstance(val, NodeMetaclass):
            return types.validate_value_type(val, value_type)
        else:
            return types.validate_data_type(value_type)

    @classmethod
    def _typed_classes(cls, *args, **kwargs):
        return _typed_scalar_exprs

    @classmethod
    def _base_class(cls, *args, **kwargs):
        return Scalar

    @classmethod
    def _transform(cls, *args, **kwargs):
        value = args[0] if len(args) > 0 else None
        value_type = args[1] if len(args) > 1 else None

        if ('_value' not in kwargs or kwargs['_value'] is None) and \
                value is not None:
            kwargs['_value'] = value
        if ('_value_type' not in kwargs or kwargs['_value_type'] is None) and \
                value_type is not None:
            kwargs['_value_type'] = types.validate_data_type(value_type)

        if kwargs.get('_value') is not None:
            kwargs['_value_type'] = types.validate_value_type(kwargs.get('_value'),
                                                              kwargs.get('_value_type'))

        if '_source_name' not in kwargs:
            kwargs['_source_name'] = kwargs.get('_name')

        if '_source_value_type' in kwargs:
            kwargs['_source_value_type'] = types.validate_data_type(kwargs['_source_value_type'])
        else:
            kwargs['_source_value_type'] = kwargs['_value_type']
        return kwargs

    def __new__(cls, *args, **kwargs):
        kwargs = cls._transform(*args, **kwargs)
        return super(Scalar, cls).__new__(cls, **kwargs)

    def _init(self, *args, **kwargs):
        self._init_attr('_name', None)
        self._init_attr('_value', None)

        kwargs = self._transform(*args, **kwargs)
        super(Scalar, self)._init(**kwargs)

    def equals(self, other):
        return super(Scalar, self).equals(other)

    @property
    def value(self):
        return self._value

    @property
    def value_type(self):
        return self._value_type

    @property
    def dtype(self):
        return self._value_type

    def output_type(self):
        return 'Scalar[%s]' % repr(self._value_type)

    def astype(self, value_type):
        value_type = types.validate_data_type(value_type)

        if value_type == self._value_type:
            return self

        attr_dict = dict()

        attr_dict['_input'] = self
        attr_dict['_value_type'] = value_type
        attr_dict['_source_value_type'] = self._source_value_type

        new_scalar = AsTypedScalar(**attr_dict)

        return new_scalar

    def to_sequence(self):
        if self._value is None:
            attr_values = dict((attr, getattr(self, attr)) for attr in get_attrs(self))

            attr_values['_data_type'] = attr_values.pop('_value_type')
            if '_source_value_type' in attr_values:
                attr_values['_source_data_type'] = attr_values.pop('_source_value_type')
            del attr_values['_value']

            cls = next(c for c in inspect.getmro(type(self))[1:]
                       if c.__name__ == type(self).__name__ and not issubclass(c, Scalar))
            seq = cls._new(**attr_values)
            return seq

        raise ExpressionError('Cannot convert valued scalar to sequence')

    def accept(self, visitor):
        visitor.visit_scalar(self)

    def get_cached(self, data):
        return Scalar(_value=data, _value_type=self.dtype)


class AsTypedScalar(Scalar):
    _args = '_input',
    node_name = "TypedScalar"

    def accept(self, visitor):
        return visitor.visit_cast(self)

    @property
    def name(self):
        return self._name or self._input.name

    @property
    def source_name(self):
        return self._source_name or self._input.source_name

    @property
    def value(self):
        return self._value

    @property
    def value_type(self):
        return self._value_type

    @property
    def dtype(self):
        return self._value_type

    def rebuild(self):
        attr_dict = self._attr_dict()
        tp = self._copy_type()
        if isinstance(attr_dict['_source_value_type'], types.Unknown):
            attr_dict['_source_value_type'] = self._input.dtype
        return tp(**attr_dict)

    @property
    def source_type(self):
        return self._source_value_type


class BooleanScalar(Scalar):
    def _init(self, *args, **kwargs):
        super(BooleanScalar, self)._init(*args, **kwargs)
        self._value_type = types.boolean


class Int8Scalar(Scalar):
    def _init(self, *args, **kwargs):
        super(Int8Scalar, self)._init(*args, **kwargs)
        self._value_type = types.int8


class Int16Scalar(Scalar):
    def _init(self, *args, **kwargs):
        super(Int16Scalar, self)._init(*args, **kwargs)
        self._value_type = types.int16


class Int32Scalar(Scalar):
    def _init(self, *args, **kwargs):
        super(Int32Scalar, self)._init(*args, **kwargs)
        self._value_type = types.int32


class Int64Scalar(Scalar):
    def _init(self, *args, **kwargs):
        super(Int64Scalar, self)._init(*args, **kwargs)
        self._value_type = types.int64


class Float32Scalar(Scalar):
    def _init(self, *args, **kwargs):
        super(Float32Scalar, self)._init(*args, **kwargs)
        self._value_type = types.float32


class Float64Scalar(Scalar):
    def _init(self, *args, **kwargs):
        super(Float64Scalar, self)._init(*args, **kwargs)
        self._value_type = types.float64


class DecimalScalar(Scalar):
    def _init(self, *args, **kwargs):
        super(DecimalScalar, self)._init(*args, **kwargs)
        self._value_type = types.decimal


class StringScalar(Scalar):
    def _init(self, *args, **kwargs):
        super(StringScalar, self)._init(*args, **kwargs)
        self._value_type = types.string


class DateScalar(Scalar):
    def _init(self, *args, **kwargs):
        super(DateScalar, self)._init(*args, **kwargs)
        self._value_type = types.date


class DatetimeScalar(Scalar):
    def _init(self, *args, **kwargs):
        super(DatetimeScalar, self)._init(*args, **kwargs)
        self._value_type = types.datetime


class TimestampScalar(Scalar):
    def _init(self, *args, **kwargs):
        super(TimestampScalar, self)._init(*args, **kwargs)
        self._value_type = types.timestamp


class BinaryScalar(Scalar):
    def _init(self, *args, **kwargs):
        super(BinaryScalar, self)._init(*args, **kwargs)
        self._value_type = types.binary


class JsonScalar(Scalar):
    def _init(self, *args, **kwargs):
        super(JsonScalar, self)._init(*args, **kwargs)
        self._value_type = types.json


class ListScalar(Scalar):
    pass


class DictScalar(Scalar):
    pass


class StructScalar(Scalar):
    pass


class UnknownScalar(Scalar):
    def _init(self, *args, **kwargs):
        super(UnknownScalar, self)._init(*args, **kwargs)
        if not isinstance(self._value_type, types.Unknown):
            self._value_type = types.Unknown()


_typed_scalar_exprs = [globals()[t.__class__.__name__ + Scalar.__name__]
                       for t in types._data_types.values()]

number_scalars = [globals().get(repr(t).capitalize() + Scalar.__name__)
                  for t in types.number_types()]

int_number_scalars = [globals().get(repr(t).capitalize() + Scalar.__name__)
                      for t in types.number_types() if repr(t).startswith('int')]


class BuiltinFunction(Scalar):
    __slots__ = '_func_name', '_func_args', '_func_kwargs'

    def __init__(self, name=None, rtype=None, args=(), **kwargs):
        rtype = rtype or kwargs.pop('_value_type', types.string)
        rtype = types.validate_data_type(rtype)
        func_name = name or kwargs.pop('_func_name', None)
        func_args = args or kwargs.pop('_func_args', ())
        super(BuiltinFunction, self).__init__(_func_name=func_name,
                                              _func_args=func_args,
                                              _value_type=rtype,
                                              **kwargs)

    def accept(self, visitor):
        visitor.visit_builtin_function(self)


class RandomScalar(BuiltinFunction):
    """
    Represent for the random scalar type.

    :param seed: random seed, None by default

    :Example:
    >>> df[df, RandomScalar().rename('append_random')]
    """
    def __new__(cls, seed=None, **kw):
        args = (seed, ) if seed is not None else kw.get('_func_args', ())
        kw.update(dict(_func_name='rand', _value_type='float', _func_args=args))
        return BuiltinFunction.__new__(cls, **kw)

    def __init__(self, seed=None, **kw):
        args = (seed, ) if seed is not None else kw.get('_func_args', ())
        kw.update(dict(_func_name='rand', _value_type='float', _func_args=args))
        super(RandomScalar, self).__init__(**kw)


class FilterCollectionExpr(CollectionExpr):
    _args = '_input', '_predicate'
    node_name = 'Filter'

    def _init(self, *args, **kwargs):
        super(FilterCollectionExpr, self)._init(*args, **kwargs)

        if self._schema is None:
            self._schema = self._input.schema

    def iter_args(self):
        for it in zip(['collection', 'predicate'], self.args):
            yield it

    @property
    def input(self):
        return self._input

    def rebuild(self):
        rebuilt = super(FilterCollectionExpr, self).rebuild()
        rebuilt._schema = rebuilt.input.schema
        return rebuilt

    def accept(self, visitor):
        visitor.visit_filter_collection(self)


class ProjectCollectionExpr(CollectionExpr):
    __slots__ = '_raw_fields',
    _args = '_input', '_fields'
    _extra_args = '_raw_fields',
    node_name = 'Projection'

    def _init(self, *args, **kwargs):
        fields = kwargs.get('_fields')
        if fields is None and len(args) >= 2:
            fields = args[1]
        for field in fields:
            if field.name is None:
                raise ExpressionError('Column does not have a name, '
                                      'please specify one by `rename`: %s' % repr_obj(field._repr()))

        self._init_attr('_raw_fields', None)
        super(ProjectCollectionExpr, self)._init(*args, **kwargs)

    def _set_field(self, value, value_dag=None):
        from ..backends.context import context
        from .window import Window

        if context.is_cached(self):
            super(ProjectCollectionExpr, self)._set_field(value, value_dag=value_dag)
            return

        if value_dag is None:
            value_dag = value.to_dag(copy=False, validate=False)

        for n in value.traverse(top_down=True, unique=True,
                                stop_cond=lambda x: isinstance(x, Column)):
            if isinstance(n, Column) and (n.input is self or n.input._proxy is self):
                source_name = n.source_name
                idx = self._schema._name_indexes[source_name]
                field = self._fields[idx]
                if field.name != n.name:
                    field = field.rename(n.name)
                value_dag.substitute(n, field)
            elif isinstance(n, Window) and n.input is self:
                # Window object like rank will point to collection directly instead of through column
                n._input = self.input

        value = value_dag.root
        fields = [field if field.name != value.name else value for field in self._fields]
        if value.name not in self._schema:
            fields.append(value)

        self._schema = TableSchema.from_lists(
            [f.name for f in fields], [f.dtype for f in fields]
        )
        self._fields = fields

    def _delitem(self, key):
        from ..backends.context import context

        if context.is_cached(self):
            return super(ProjectCollectionExpr, self).__delitem__(key)

        fields = [f for f in self._fields if f.name != key]
        self._schema = TableSchema.from_lists(
            [f.name for f in fields], [f.dtype for f in fields]
        )
        self._fields = fields

    def __delitem__(self, key):
        if key not in self._schema:
            raise KeyError('Field({0}) does not exist'.format(key))

        self._delitem(key)

    @property
    def _project_fields(self):
        return self._fields

    def iter_args(self):
        for it in zip(['collection', 'selections'], self.args):
            yield it

    @property
    def input(self):
        return self._input

    @property
    def fields(self):
        return self._fields

    def rebuild(self):
        if self._raw_fields:
            return self._input.select(*self._raw_fields)
        rebuilt = super(ProjectCollectionExpr, self).rebuild()
        rebuilt._schema = TableSchema.from_lists(
            [f.name for f in rebuilt._fields], [f.dtype for f in rebuilt._fields]
        )
        return rebuilt

    def accept(self, visitor):
        visitor.visit_project_collection(self)


class LateralViewCollectionExpr(ProjectCollectionExpr):
    _args = '_input', '_fields', '_lateral_views'
    node_name = 'LateralView'

    def _init(self, *args, **kwargs):
        self._init_attr('_lateral_views', None)
        super(LateralViewCollectionExpr, self)._init(*args, **kwargs)
        self._redirect_lateral_views()

    def _redirect_lateral_views(self):
        # dealing with cases like df[df['a', 'b'].apply(..., axis=1), df.c]
        for lv in self.lateral_views:
            if lv.input is self.input:
                continue
            for f, f_o in zip(lv._fields, lv.input._fields):
                lv.substitute(f, f_o)
            lv.substitute(lv.input, self.input)

    def iter_args(self):
        for it in zip(['collection', 'selections', 'lateral_views'], self.args):
            yield it

    @property
    def lateral_views(self):
        return self._lateral_views

    def accept(self, visitor):
        visitor.visit_lateral_view(self)


class FilterPartitionCollectionExpr(CollectionExpr):
    __slots__ = '_predicate_string',

    _args = '_input', '_predicate', '_fields'
    node_name = 'FilterPartition'

    def _init(self, *args, **kwargs):
        super(FilterPartitionCollectionExpr, self)._init(*args, **kwargs)
        self._fields = [self._input[col.name] for col in self._schema.columns]
        self._predicate_string = kwargs.get('_predicate_string')

    @property
    def _project_fields(self):
        return self._fields

    def iter_args(self):
        for it in zip(['collection', 'predicate', 'selections'], self.args):
            yield it

    @property
    def input(self):
        return self._input

    @property
    def fields(self):
        return self._fields

    @property
    def predicate_string(self):
        return self._predicate_string

    def accept(self, visitor):
        visitor.visit_filter_partition_collection(self)

    @property
    def data_table(self):
        return self.input.data

    @property
    def data(self):
        from ...types import PartitionSpec
        from .arithmetic import Or

        if isinstance(self._predicate, Or):
            raise AttributeError('Cannot get data when predicate contains multiple partition specs.')
        spec = PartitionSpec(self.predicate_string)
        return self.data_table.partitions[spec]


class SliceCollectionExpr(CollectionExpr):

    _args = '_input', '_indexes'
    node_name = 'Slice'

    def _init(self, *args, **kwargs):
        super(SliceCollectionExpr, self)._init(*args, **kwargs)

        if isinstance(self._indexes, slice):
            scalar = lambda v: Scalar(_value=v) if v is not None else None
            self._indexes = scalar(self._indexes.start), \
                            scalar(self._indexes.stop), scalar(self._indexes.step)

    @property
    def start(self):
        res = self._indexes[0]
        return res.value if res is not None else None

    @property
    def stop(self):
        res = self._indexes[1]
        return res.value if res is not None else None

    @property
    def step(self):
        res = self._indexes[2]
        return res.value if res is not None else None

    @property
    def input(self):
        return self._input

    def iter_args(self):
        args = [self._input] + list(self._indexes)
        for it in zip(['collection', 'start', 'stop', 'step'], args):
            yield it

    def rebuild(self):
        rebuilt = super(SliceCollectionExpr, self).rebuild()
        rebuilt._schema = self.input.schema
        return rebuilt

    def accept(self, visitor):
        visitor.visit_slice_collection(self)


class Summary(CollectionExpr):
    __slots__ = '_schema',
    _args = '_input', '_fields'

    def _init(self, *args, **kwargs):
        super(Summary, self)._init(*args, **kwargs)
        if hasattr(self, '_schema') and any(it is None for it in self._schema.names):
            raise TypeError('Schema cannot has field which name is None')

    @property
    def input(self):
        return self._input

    @property
    def fields(self):
        return self._fields

    @property
    def schema(self):
        return self._schema

    def iter_args(self):
        for it in zip(['collection', 'fields'], self.args):
            yield it

    def accept(self, visitor):
        visitor.visit_project_collection(self)


from . import element
from . import arithmetic
from . import reduction
from . import groupby
from . import collections
from . import window
from . import math
from . import strings
from . import datetimes
from . import merge
from . import composites
from ..tools import *


# hack for count
def _count(expr, *args, **kwargs):
    if len(args) + len(kwargs) > 0:
        from .strings import _count
        return _count(expr, *args, **kwargs)
    else:
        from .reduction import count
        return count(expr)


StringSequenceExpr.count = _count



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2022 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import warnings

from .. import serializers, errors, utils
from ..compat import six, Enum
from ..errors import InternalServerError
from .cache import cache
from .core import Iterable, LazyLoad


class Volume(LazyLoad):
    """
    Volume is the file-accessing object provided by ODPS.
    """

    EXTERNAL_VOLUME_LOCATION_KEY = "external.location"
    EXTERNAL_VOLUME_ROLEARN_KEY = "odps.properties.rolearn"

    class Type(Enum):
        NEW = 'NEW'
        OLD = 'OLD'
        EXTERNAL = 'EXTERNAL'
        UNKNOWN = 'UNKNOWN'

    _root = 'Meta'
    _type_indicator = 'type'

    name = serializers.XMLNodeField('Name')
    owner = serializers.XMLNodeField('Owner')
    comment = serializers.XMLNodeField('Comment')
    type = serializers.XMLNodeField(
        'Type', parse_callback=lambda t: Volume.Type(t.upper()), serialize_callback=lambda t: t.value
    )
    length = serializers.XMLNodeField('Length', parse_callback=int)
    file_number = serializers.XMLNodeField('FileNumber', parse_callback=int)
    creation_time = serializers.XMLNodeField('CreationTime', parse_callback=utils.parse_rfc822)
    last_modified_time = serializers.XMLNodeField('LastModifiedTime', parse_callback=utils.parse_rfc822)
    properties = serializers.XMLNodeField(
        'Properties', parse_callback=json.loads, serialize_callback=json.dumps
    )

    @classmethod
    def _get_cls(cls, typo):
        if typo is None:
            return cls
        if isinstance(typo, six.string_types):
            typo = Volume.Type(typo.upper())

        if typo == Volume.Type.OLD:
            from . import PartedVolume
            return PartedVolume
        elif typo == Volume.Type.NEW:
            from . import FSVolume
            return FSVolume
        elif typo == Volume.Type.EXTERNAL:
            from . import ExternalVolume
            return ExternalVolume
        elif typo == Volume.Type.UNKNOWN:
            return Volume

    @staticmethod
    def _filter_cache(_, **kwargs):
        return kwargs.get('type') is not None and kwargs['type'] != Volume.Type.UNKNOWN

    @cache
    def __new__(cls, *args, **kwargs):
        typo = kwargs.get('type')
        if typo is not None or (cls != Volume and issubclass(cls, Volume)):
            return object.__new__(cls._get_cls(typo))

        kwargs['type'] = Volume.Type.UNKNOWN
        obj = Volume(**kwargs)
        try:
            obj.reload()
            return Volume(**obj.extract())
        except InternalServerError as ex:
            warnings.warn(
                "Cannot reload volume %s due to error %s" % (obj.name, str(ex))
            )
            return obj

    def __init__(self, **kwargs):
        typo = kwargs.get('type')

        properties = kwargs.get("properties") or {}
        location = kwargs.pop("location", None)
        rolearn = kwargs.pop("rolearn", None)
        if location:
            properties[self.EXTERNAL_VOLUME_LOCATION_KEY] = location
        if rolearn:
            properties[self.EXTERNAL_VOLUME_ROLEARN_KEY] = rolearn
        if properties:
            kwargs["properties"] = properties

        if isinstance(typo, six.string_types):
            kwargs['type'] = Volume.Type(typo.upper())
        super(Volume, self).__init__(**kwargs)

    def reload(self):
        params = {}
        schema_name = self._get_schema_name()
        if schema_name is not None:
            params['curr_schema'] = schema_name
        resp = self._client.get(self.resource(), action='meta', params=params)
        self.parse(self._client, resp, obj=self)

    def drop(self):
        return self.parent.delete(self)


class Volumes(Iterable):

    marker = serializers.XMLNodeField('Marker')
    volumes = serializers.XMLNodesReferencesField(Volume, 'Volume')

    def _get(self, item):
        return Volume(client=self._client, parent=self, name=item)

    def __contains__(self, item):
        if isinstance(item, six.string_types):
            try:
                # as reload() will be done in constructor of Volume, we return directly.
                return self._get(item)
            except errors.NoSuchObject:
                return False
        elif isinstance(item, Volume):
            volume = item
            try:
                volume.reload()
                return True
            except errors.NoSuchObject:
                return False
        else:
            return False

    def __iter__(self):
        return self.iterate()

    def iterate(self, name=None, owner=None):
        """

        :param name: the prefix of volume name name
        :param owner:
        :return:
        """
        schema_name = self._get_schema_name()
        params = {'expectmarker': 'true'}
        if name is not None:
            params['name'] = name
        if owner is not None:
            params['owner'] = owner
        if schema_name is not None:
            params['curr_schema'] = schema_name

        def _it():
            last_marker = params.get('marker')
            if 'marker' in params and \
                    (last_marker is None or len(last_marker) == 0):
                return

            url = self.resource()
            resp = self._client.get(url, params=params)

            v = Volumes.parse(self._client, resp, obj=self)
            params['marker'] = v.marker

            return v.volumes

        while True:
            volumes = _it()
            if volumes is None:
                break
            for volume in volumes:
                yield volume

    def _create(self, obj=None, **kwargs):
        volume = obj or Volume(parent=self, client=self._client, **kwargs)
        if volume.parent is None:
            volume._parent = self
        if volume._client is None:
            volume._client = self._client

        headers = {'Content-Type': 'application/xml'}
        data = volume.serialize()

        self._client.post(
            self.resource(), data, headers=headers, curr_schema=self._get_schema_name()
        )
        return self[volume.name]

    def create_parted(self, obj=None, **kwargs):
        return self._create(obj=obj, type='old', **kwargs)

    def create_fs(self, obj=None, **kwargs):
        return self._create(obj=obj, type='new', **kwargs)

    def create_external(self, obj=None, **kwargs):
        return self._create(obj=obj, type='external', **kwargs)

    def delete(self, name):
        if not isinstance(name, Volume):
            volume = Volume(name=name, parent=self, client=self._client)
        else:
            volume = name
            name = name.name
        del self[name]  # release cache

        url = volume.resource()

        self._client.delete(url, curr_schema=self._get_schema_name())



# encoding: utf-8
# Copyright 1999-2022 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import glob
import codecs
import shutil
import tarfile
import gzip
import warnings
from collections import OrderedDict

from itertools import groupby, product

from ..compat import pickle, urlretrieve, six
from ..tunnel import TableTunnel
from ..utils import load_static_text_file, build_pyodps_dir

USER_DATA_REPO = build_pyodps_dir('data')


class TestDataMixIn(object):
    def after_create_test_data(self, table_name):
        pass


def table_creator(func):
    """
    Decorator for table creating method
    """
    @six.wraps(func)
    def method(self, table_name, **kwargs):
        if self.odps.exist_table(table_name):
            table = self.odps.get_table(table_name)
            if list(table.head(1)):
                return
            self.odps.delete_table(table_name)
        if kwargs.get('project', self.odps.project) != self.odps.project:
            tunnel = TableTunnel(self.odps, project=kwargs['project'])
        else:
            tunnel = self.tunnel
        func(self.odps, table_name, tunnel=tunnel, **kwargs)
        self.after_create_test_data(table_name)

    setattr(TestDataMixIn, func.__name__, method)
    return func


"""
Simple DataFrames
"""


@table_creator
def create_ionosphere(odps, table_name, tunnel=None, project=None):
    if tunnel is None:
        tunnel = TableTunnel(odps, project=project)
    fields = ','.join('a%02d double' % i for i in range(1, 35)) + ', class bigint'
    odps.delete_table(table_name, if_exists=True, project=project)
    odps.create_table(table_name, fields, project=project)

    upload_ss = tunnel.create_upload_session(table_name)
    writer = upload_ss.open_record_writer(0)

    for line in load_static_text_file('data/ionosphere.txt').splitlines():
        rec = upload_ss.new_record()
        cols = [float(c) if rec._columns[i].type == 'double' else int(c) for i, c in enumerate(line.split(','))]
        [rec.set(i, val) for i, val in enumerate(cols)]
        writer.write(rec)
    writer.close()
    upload_ss.commit([0, ])


@table_creator
def create_ionosphere_one_part(odps, table_name, partition_count=3, tunnel=None, project=None):
    if tunnel is None:
        tunnel = TableTunnel(odps, project=project)
    fields = ','.join('a%02d double' % i for i in range(1, 35)) + ', class bigint'
    odps.delete_table(table_name, if_exists=True, project=project)
    odps.create_table(table_name, (fields, 'part bigint'), project=project)
    for part_id in range(partition_count):
        odps.execute_sql('alter table %s add if not exists partition (part=%d)' % (table_name, part_id), project=project)

    upload_sses = [tunnel.create_upload_session(table_name, 'part=%d' % part_id) for part_id in range(partition_count)]
    writers = [session.open_record_writer(0) for session in upload_sses]

    for line_no, line in enumerate(load_static_text_file('data/ionosphere.txt').splitlines()):
        part_id = line_no % partition_count
        rec = upload_sses[part_id].new_record()
        cols = [float(c) if rec._columns[i].type == 'double' else int(c) for i, c in enumerate(line.split(','))]
        cols.append(part_id)
        [rec.set(i, val) for i, val in enumerate(cols)]
        writers[part_id].write(rec)
    [writer.close() for writer in writers]
    [upload_ss.commit([0, ]) for upload_ss in upload_sses]


@table_creator
def create_ionosphere_two_parts(odps, table_name, partition1_count=2, partition2_count=3, tunnel=None, project=None):
    if tunnel is None:
        tunnel = TableTunnel(odps, project=project)
    fields = ','.join('a%02d double' % i for i in range(1, 35)) + ', class bigint'
    odps.delete_table(table_name, if_exists=True, project=project)
    odps.create_table(table_name, (fields, 'part1 bigint, part2 bigint'), project=project)
    for id1, id2 in product(range(partition1_count), range(partition2_count)):
        odps.execute_sql('alter table %s add if not exists partition (part1=%d, part2=%d)' % (table_name, id1, id2),
                         project=project)

    upload_sses = [[tunnel.create_upload_session(table_name, 'part1=%d,part2=%d' % (id1, id2))
                    for id2 in range(partition2_count)] for id1 in range(partition1_count)]
    writers = [[session.open_record_writer(0) for session in sessions] for sessions in upload_sses]

    for line_no, line in enumerate(load_static_text_file('data/ionosphere.txt').splitlines()):
        id1, id2 = line_no % partition1_count, line_no % partition2_count
        rec = upload_sses[id1][id2].new_record()
        cols = [float(c) if rec._columns[i].type == 'double' else int(c) for i, c in enumerate(line.split(','))]
        cols.extend([id1, id2])
        [rec.set(i, val) for i, val in enumerate(cols)]
        writers[id1][id2].write(rec)
    [writer.close() for ws in writers for writer in ws]
    [upload_ss.commit([0, ]) for upload_sss in upload_sses for upload_ss in upload_sss]


@table_creator
def create_iris(odps, table_name, tunnel=None, project=None, lifecycle=None):
    if tunnel is None:
        tunnel = TableTunnel(odps, project=project)
    odps.delete_table(table_name, if_exists=True, project=project)
    table = odps.create_table(table_name, 'sepal_length double, sepal_width double, petal_length double, '
                              + 'petal_width double, category string', project=project, lifecycle=lifecycle)

    upload_ss = tunnel.create_upload_session(table_name)
    writer = upload_ss.open_record_writer(0)

    for line in load_static_text_file('data/iris.txt').splitlines():
        rec = upload_ss.new_record()
        line_parts = line.split(',')
        cols = [float(c) for c in line_parts[:-1]]
        cols.append(line_parts[4])
        [rec.set(i, val) for i, val in enumerate(cols)]
        writer.write(rec)
    writer.close()
    upload_ss.commit([0, ])
    return table


@table_creator
def create_iris_kv(odps, table_name, tunnel=None, project=None):
    if tunnel is None:
        tunnel = TableTunnel(odps, project=project)
    odps.delete_table(table_name, if_exists=True, project=project)
    odps.create_table(table_name, 'content string, category bigint', project=project)

    upload_ss = tunnel.create_upload_session(table_name)
    writer = upload_ss.open_record_writer(0)

    for line in load_static_text_file('data/iris.txt').splitlines():
        rec = upload_ss.new_record()
        line_parts = line.split(',')
        rec.set(0, ','.join('%s:%s' % (idx, c) for idx, c in enumerate(line_parts[:-1])))
        rec.set(1, 0 if 'setosa' in line_parts[-1] else 1)
        writer.write(rec)
    writer.close()
    upload_ss.commit([0, ])


@table_creator
def create_corpus(odps, table_name, tunnel=None, project=None):
    if tunnel is None:
        tunnel = TableTunnel(odps, project=project)
    odps.delete_table(table_name, if_exists=True, project=project)
    odps.create_table(table_name, 'id string, content string', project=project)

    upload_ss = tunnel.create_upload_session(table_name)
    writer = upload_ss.open_record_writer(0)

    for line_no, line in enumerate(load_static_text_file('data/splited_words.txt').splitlines()):
        rec = upload_ss.new_record()
        cols = [line_no + 1, line.replace('####', '')]
        [rec.set(i, val) for i, val in enumerate(cols)]
        writer.write(rec)
    writer.close()
    upload_ss.commit([0, ])


@table_creator
def create_word_triple(odps, table_name, tunnel=None, project=None):
    if tunnel is None:
        tunnel = TableTunnel(odps, project=project)
    odps.delete_table(table_name, if_exists=True, project=project)
    odps.create_table(table_name, 'id string, word string, count bigint', project=project)

    upload_ss = tunnel.create_upload_session(table_name)
    writer = upload_ss.open_record_writer(0)

    for line_no, line in enumerate(load_static_text_file('data/splited_words.txt').splitlines()):
        line = line.strip()
        if not line:
            break
        for word, group in groupby(sorted(line.split('####'))):
            rec = upload_ss.new_record()
            cols = [str(line_no + 1), word, len(list(group))]
            [rec.set(i, val) for i, val in enumerate(cols)]
            writer.write(rec)
    writer.close()
    upload_ss.commit([0, ])


@table_creator
def create_splited_words(odps, table_name, joined=False, tunnel=None, project=None):
    if tunnel is None:
        tunnel = TableTunnel(odps, project=project)
    odps.delete_table(table_name, if_exists=True, project=project)
    odps.create_table(table_name, 'doc_id string, content string', project=project)

    upload_ss = tunnel.create_upload_session(table_name)
    writer = upload_ss.open_record_writer(0)

    for line_no, line in enumerate(load_static_text_file('data/splited_words.txt').splitlines()):
        if not line.strip():
            break
        if joined:
            rec = upload_ss.new_record()
            cols = [line_no + 1, line.replace('####', ' ')]
            [rec.set(i, val) for i, val in enumerate(cols)]
            writer.write(rec)
        else:
            for word in line.split('####'):
                if not word:
                    continue
                rec = upload_ss.new_record()
                cols = [line_no + 1, word]
                [rec.set(i, val) for i, val in enumerate(cols)]
                writer.write(rec)
    writer.close()
    upload_ss.commit([0, ])


@table_creator
def create_weighted_graph_edges(odps, table_name, tunnel=None, project=None):
    if tunnel is None:
        tunnel = TableTunnel(odps, project=project)
    data_rows = [
        ['1', '1', '2', '1', 0.7], ['1', '1', '3', '1', 0.7], ['1', '1', '4', '1', 0.6], ['2', '1', '3', '1', 0.7],
        ['2', '1', '4', '1', 0.6], ['3', '1', '4', '1', 0.6], ['4', '1', '6', '5', 0.3], ['5', '5', '6', '5', 0.6],
        ['5', '5', '7', '5', 0.7], ['5', '5', '8', '5', 0.7], ['6', '5', '7', '5', 0.6], ['6', '5', '8', '5', 0.6],
        ['7', '5', '8', '5', 0.7]
    ]
    odps.delete_table(table_name, if_exists=True, project=project)
    odps.create_table(table_name, 'flow_out_id string, group_out_id string, flow_in_id string, ' +
                      'group_in_id string, edge_weight double', project=project)

    upload_ss = tunnel.create_upload_session(table_name)
    writer = upload_ss.open_record_writer(0)

    for rd in data_rows:
        rec = upload_ss.new_record()
        [rec.set(i, val) for i, val in enumerate(rd)]
        writer.write(rec)
    writer.close()
    upload_ss.commit([0, ])


@table_creator
def create_tree_graph(odps, table_name, tunnel=None, project=None):
    if tunnel is None:
        tunnel = TableTunnel(odps, project=project)
    data_rows = [
        ['0', '1'], ['0', '2'], ['1', '3'], ['1', '4'], ['2', '4'], ['2', '5'], ['4', '6'], ['a', 'b'], ['a', 'c'],
        ['c', 'd'], ['c', 'e']
    ]
    odps.delete_table(table_name, if_exists=True, project=project)
    odps.create_table(table_name, 'flow_out_id string, flow_in_id string', project=project)

    upload_ss = tunnel.create_upload_session(table_name)
    writer = upload_ss.open_record_writer(0)

    for rd in data_rows:
        rec = upload_ss.new_record()
        [rec.set(i, val) for i, val in enumerate(rd)]
        writer.write(rec)
    writer.close()
    upload_ss.commit([0, ])


@table_creator
def create_weighted_graph_vertices(odps, table_name, tunnel=None, project=None):
    if tunnel is None:
        tunnel = TableTunnel(odps, project=project)
    data_rows = [
        ['1', '1', 0.7, 1.0], ['2', '1', 0.7, 1.0], ['3', '1', 0.7, 1.0], ['4', '1', 0.5, 1.0], ['5', '5', 0.7, 1.0],
        ['6', '5', 0.5, 1.0], ['7', '5', 0.7, 1.0], ['8', '5', 0.7, 1.0]
    ]
    odps.delete_table(table_name, if_exists=True, project=project)
    odps.create_table(table_name, 'node string, label string, node_weight double, label_weight double',
                      project=project)

    upload_ss = tunnel.create_upload_session(table_name)
    writer = upload_ss.open_record_writer(0)

    for rd in data_rows:
        rec = upload_ss.new_record()
        [rec.set(i, val) for i, val in enumerate(rd)]
        writer.write(rec)
    writer.close()
    upload_ss.commit([0, ])


@table_creator
def create_user_item_table(odps, table_name, tunnel=None, mode=None, project=None):
    if tunnel is None:
        tunnel = TableTunnel(odps, project=project)
    data_rows = [
        ['CST0000', 'a', 0], ['CST0000', 'b', 1], ['CST0000', 'c', 2], ['CST0000', 'd', 3], ['CST0001', 'a', 0],
        ['CST0001', 'b', 0], ['CST0001', 'a', 1], ['CST0001', 'b', 1], ['CST0001', 'c', 1], ['CST0001', 'b', 2],
        ['CST0001', 'c', 2], ['CST0001', 'd', 2], ['CST0001', 'e', 3], ['CST0002', 'a', 0], ['CST0002', 'c', 0],
        ['CST0002', 'b', 1], ['CST0002', 'a', 2], ['CST0002', 'b', 2], ['CST0002', 'c', 2], ['CST0002', 'a', 3],
        ['CST0000', 'a', 0], ['CST0000', 'b', 1], ['CST0000', 'c', 2], ['CST0000', 'd', 3], ['CST0001', 'a', 0],
        ['CST0001', 'b', 0], ['CST0001', 'a', 1], ['CST0001', 'b', 1], ['CST0001', 'c', 1], ['CST0001', 'b', 2],
        ['CST0001', 'c', 2], ['CST0001', 'd', 2], ['CST0001', 'e', 3], ['CST0002', 'a', 0], ['CST0002', 'c', 0],
        ['CST0002', 'b', 1], ['CST0002', 'a', 2], ['CST0002', 'b', 2], ['CST0002', 'c', 2], ['CST0002', 'a', 3],
        ['CST0000', 'a', 0], ['CST0000', 'b', 1], ['CST0000', 'c', 2], ['CST0000', 'd', 3], ['CST0001', 'a', 0],
        ['CST0001', 'b', 0], ['CST0001', 'a', 1], ['CST0001', 'b', 1], ['CST0001', 'c', 1], ['CST0001', 'b', 2],
        ['CST0001', 'c', 2], ['CST0001', 'd', 2], ['CST0001', 'e', 3], ['CST0002', 'a', 0], ['CST0002', 'c', 0],
        ['CST0002', 'b', 1], ['CST0002', 'a', 2], ['CST0002', 'b', 2], ['CST0002', 'c', 2], ['CST0002', 'a', 3],
        ['CST0000', 'a', 0], ['CST0000', 'b', 1], ['CST0000', 'c', 2], ['CST0000', 'd', 3], ['CST0001', 'a', 0],
        ['CST0001', 'b', 0], ['CST0001', 'a', 1], ['CST0001', 'b', 1], ['CST0001', 'c', 1], ['CST0001', 'b', 2],
        ['CST0001', 'c', 2], ['CST0001', 'd', 2], ['CST0001', 'e', 3], ['CST0002', 'a', 0], ['CST0002', 'c', 0],
        ['CST0002', 'b', 1], ['CST0002', 'a', 2], ['CST0002', 'b', 2], ['CST0002', 'c', 2], ['CST0002', 'a', 3]
    ]
    odps.delete_table(table_name, if_exists=True, project=project)
    if mode == 'agg':
        data_rows = [k + (len(list(p)), ) for k, p in groupby(sorted(data_rows, key=lambda item: (item[0], item[1])), lambda item: (item[0], item[1]))]
        odps.create_table(table_name, 'user string, item string, payload bigint', project=project)
    elif mode == 'exist':
        items = dict()
        for k, g1 in groupby(sorted(data_rows, key=lambda it: it[0]), key=lambda it: it[0]):
            items[k] = {k2: 1 for k2 in set(it2[1] for it2 in g1)}
        products = set(it[1] for it in data_rows)
        for k, g in six.iteritems(items):
            unexist_set = products - set(six.iterkeys(g))
            for it in unexist_set:
                g[it] = 0
        data_rows = [(u, p, l) for u, ud in six.iteritems(items) for p, l in six.iteritems(ud)]
        odps.create_table(table_name, 'user string, item string, label bigint', project=project)
    elif mode == 'kv':
        data_rows = [(k[0], '{0}:{1}'.format(k[1], len(list(p))))
                     for k, p in groupby(sorted(data_rows, key=lambda item: (item[0], item[1])), lambda item: (item[0], item[1]))]
        data_rows = [(k, ','.join((v[1] for v in p))) for k, p in groupby(data_rows, key=lambda item: item[0])]
        odps.create_table(table_name, 'user string, item_list string', project=project)
    else:
        odps.create_table(table_name, 'user string, item string, time bigint', project=project)

    upload_ss = tunnel.create_upload_session(table_name)
    writer = upload_ss.open_record_writer(0)

    for rd in data_rows:
        rec = upload_ss.new_record()
        [rec.set(i, val) for i, val in enumerate(rd)]
        writer.write(rec)
    writer.close()
    upload_ss.commit([0, ])


@table_creator
def create_dow_jones(odps, table_name, tunnel=None, project=None):
    fields = """quarter stock date open high low close volume percent_change_price percent_change_volume_over_last_wk
    previous_weeks_volume next_weeks_open next_weeks_close percent_change_next_weeks_price days_to_next_dividend
    percent_return_next_dividend""".strip().split()
    field_types = OrderedDict([(fname, 'double') for fname in fields])
    field_types['quarter'] = 'bigint'
    field_types['stock'] = 'string'
    field_types['date'] = 'string'
    fields_str = ','.join('`{0}` {1}'.format(k, v) for k, v in six.iteritems(field_types))

    odps.delete_table(table_name, if_exists=True, project=project)
    odps.create_table(table_name, fields_str, project=project)

    def iter_lines():
        for line in load_static_text_file('data/dow_jones.txt').splitlines():
            ldata = line.split(',')
            ldata[0] = int(ldata[0])
            ldata[3:] = [float(v) if v else None for v in ldata[3:]]
            yield ldata

    odps.write_table(table_name, iter_lines())


"""
20 Newsgroups
"""

NEWSGROUP_URL = 'http://repo.aliyun.com/shared_data/20news-bydate.tar.gz'
NEWSGROUP_DATA_NAME = '20news-bydate'
NEWSGROUP_ARCHIVE_NAME = '20news-bydate.tar.gz'
NEWSGROUP_CACHE_NAME = '20news-bydate.pkz'
NEWSGROUP_TRAIN_DIR = "20news-bydate-train"
NEWSGROUP_TEST_DIR = "20news-bydate-test"
NEWSGROUP_TRAIN_FOLDER = "20news-bydate-train"
NEWSGROUP_TEST_FOLDER = "20news-bydate-test"


def download_newsgroup(target_dir, cache_dir):
    target_tar = os.path.join(os.path.expanduser('~'), NEWSGROUP_ARCHIVE_NAME)
    urlretrieve(NEWSGROUP_URL, target_tar)
    cache_newsgroup_tar(target_tar, target_dir, cache_dir)


def cache_newsgroup_tar(target_tar, target_dir, cache_dir):
    tarfile.open(target_tar, 'r:gz').extractall(path=target_dir)
    os.unlink(target_tar)

    if not os.path.exists(target_dir):
        os.makedirs(target_dir)

    train_path = os.path.join(target_dir, NEWSGROUP_TRAIN_FOLDER)
    test_path = os.path.join(target_dir, NEWSGROUP_TEST_FOLDER)
    cache_path = os.path.join(cache_dir, NEWSGROUP_CACHE_NAME)

    if not os.path.exists(cache_dir):
        os.makedirs(cache_dir)

    def load_files(path, encoding):
        objs = []
        for fn in glob.glob(os.path.join(path, '*')):
            file_cat = os.path.basename(os.path.normpath(fn))
            for sfn in glob.glob(os.path.join(fn, '*')):
                file_id = os.path.basename(os.path.normpath(sfn))
                with open(sfn, 'rb') as f:
                    objs.append((file_id, file_cat, f.read().decode(encoding)))

        return objs

    # Store a zipped pickle
    cache = dict(train=load_files(train_path, encoding='latin1'),
                 test=load_files(test_path, encoding='latin1'))
    compressed_content = codecs.encode(pickle.dumps(cache), 'zlib_codec')
    with open(cache_path, 'wb') as f:
        f.write(compressed_content)

    shutil.rmtree(target_dir)


@table_creator
def create_newsgroup_table(odps, table_name, tunnel=None, data_part='train', project=None):
    cache_file = os.path.join(USER_DATA_REPO, NEWSGROUP_CACHE_NAME)
    if not os.path.exists(USER_DATA_REPO):
        os.makedirs(USER_DATA_REPO)
    if not os.path.exists(cache_file):
        warnings.warn('We need to download data from ' + NEWSGROUP_URL + '.')
        download_newsgroup(os.path.join(USER_DATA_REPO, NEWSGROUP_DATA_NAME), USER_DATA_REPO)

    with open(cache_file, 'rb') as f:
        cache = pickle.loads(codecs.decode(f.read(), 'zlib_codec'))

    if tunnel is None:
        tunnel = TableTunnel(odps, project=project)

    odps.delete_table(table_name, if_exists=True, project=project)
    odps.create_table(table_name, 'id string, category string, message string', project=project)

    upload_ss = tunnel.create_upload_session(table_name)
    writer = upload_ss.open_record_writer(0)

    for line in cache[data_part]:
        rec = upload_ss.new_record()
        [rec.set(i, six.text_type(val)) for i, val in enumerate(line)]
        writer.write(rec)
    writer.close()
    upload_ss.commit([0, ])


"""
MNIST
"""

MNIST_PICKLED_URL = 'http://repo.aliyun.com/shared_data/mnist.pkl.gz'
MNIST_FILE = 'mnist.pkl.gz'
mnist_unpickled = None


def load_mnist_data():
    global mnist_unpickled

    if mnist_unpickled is not None:
        return mnist_unpickled

    mnist_file = os.path.join(USER_DATA_REPO, MNIST_FILE)
    if not os.path.exists(USER_DATA_REPO):
        os.makedirs(USER_DATA_REPO)
    if not os.path.exists(mnist_file):
        warnings.warn('We need to download data from ' + MNIST_PICKLED_URL + '.')
        urlretrieve(MNIST_PICKLED_URL, mnist_file)

    with gzip.open(mnist_file, 'rb') as fobj:
        mnist_unpickled = pickle.load(fobj)
        fobj.close()

    return mnist_unpickled


@table_creator
def create_mnist_table(odps, table_name, part_id=0, tunnel=None, project=None):
    train_data = load_mnist_data()[part_id]

    if tunnel is None:
        tunnel = TableTunnel(odps)

    odps.delete_table(table_name, if_exists=True, project=project)
    odps.create_table(table_name, 'feature string, class string')

    upload_ss = tunnel.create_upload_session(table_name)
    writer = upload_ss.open_record_writer(0)

    for feature, label in zip(train_data[0], train_data[1]):
        rec = upload_ss.new_record()
        rec.set(0, six.text_type(','.join(str(n) for n in feature)))
        rec.set(1, six.text_type(label))
        writer.write(rec)

    writer.close()
    upload_ss.commit([0, ])
